"""
The LXD modules manage LXD containers
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.lxd import LxdContainers


def get_container_named(name, containers):
    for container in containers:
        if container["name"] == name:
            return container
    else:
        return None


@operation
def container(
    id,
    present=True,
    image="ubuntu:16.04",
):
    """
    Add/remove LXD containers.

    Note: does not check if an existing container is based on the specified
    image.

    + id: name/identifier for the container
    + image: image to base the container on
    + present: whether the container should be present or absent

    **Example:**

    .. code:: python

        lxd.container(
            name="Add an ubuntu container",
            id="ubuntu19",
            image="ubuntu:19.10",
        )
    """

    current_containers = host.get_fact(LxdContainers)
    container = get_container_named(id, current_containers)

    # Container exists and we don't want it
    if not present:
        if container:
            if container["status"] == "Running":
                yield "lxc stop {0}".format(id)

            # Command to remove the container:
            yield "lxc delete {0}".format(id)

            current_containers.remove(container)
        else:
            host.noop("container {0} does not exist".format(id))

    # Container doesn't exist and we want it
    if present:
        if not container:
            # Command to create the container:
            yield "lxc launch {image} {id} < /dev/null".format(id=id, image=image)
            current_containers.append(
                {
                    "name": id,
                    "image": image,
                },
            )
        else:
            host.noop("container {0} exists".format(id))



"""
Manage cargo (aka Rust) packages.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.cargo import CargoPackages

from .util.packaging import ensure_packages


@operation
def packages(packages=None, present=True, latest=False):
    """
    Install/remove/update cargo packages.

    + packages: list of packages to ensure
    + present: whether the packages should be present
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like cargo: ``<pkg>@<version>``.
    """

    current_packages = host.get_fact(CargoPackages)

    install_command = "cargo install"

    uninstall_command = "cargo uninstall"

    upgrade_command = "cargo install"

    yield from ensure_packages(
        host,
        packages,
        current_packages,
        present,
        install_command=install_command,
        uninstall_command=uninstall_command,
        upgrade_command=upgrade_command,
        version_join="@",
        latest=latest,
    )



from unittest import TestCase

from pyinfra.api import Config, Inventory, State
from pyinfra.api.arguments import pop_global_arguments


class TestOperationKwargs(TestCase):
    def test_get_from_config(self):
        config = Config(SUDO="config-value")
        inventory = Inventory((("somehost",), {}))

        state = State(config=config, inventory=inventory)

        kwargs, keys = pop_global_arguments({}, state=state, host=inventory.get_host("somehost"))
        assert kwargs["sudo"] == "config-value"

    def test_get_from_host(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"sudo": "host-value"})], {}))

        state = State(config=config, inventory=inventory)

        kwargs, keys = pop_global_arguments({}, state=state, host=inventory.get_host("somehost"))
        assert kwargs["sudo"] == "host-value"

    def test_get_from_state_deploy_kwargs(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"sudo": "host-value"})], {}))
        somehost = inventory.get_host("somehost")

        state = State(config=config, inventory=inventory)
        somehost.current_deploy_kwargs = {"sudo": "deploy-kwarg-value"}

        kwargs, keys = pop_global_arguments({}, state=state, host=somehost)
        assert kwargs["sudo"] == "deploy-kwarg-value"

    def test_get_from_kwargs(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"sudo": "host-value"})], {}))
        somehost = inventory.get_host("somehost")

        state = State(config=config, inventory=inventory)
        somehost.current_deploy_kwargs = {
            "sudo": "deploy-kwarg-value",
            "sudo_user": "deploy-kwarg-user",
        }

        kwargs, keys = pop_global_arguments(
            {"sudo": "kwarg-value"},
            state=state,
            host=somehost,
        )
        assert kwargs["sudo"] == "kwarg-value"
        assert kwargs["sudo_user"] == "deploy-kwarg-user"
        assert "sudo" in keys



from inspect import isclass
from unittest import TestCase

from paramiko import RSAKey, SFTPClient, SSHClient
from paramiko.agent import AgentRequestHandler

from pyinfra.connectors import ssh


class PatchSSHTestCase(TestCase):
    """
    A test class that patches out the paramiko SSH parts such that they succeed as normal.
    The SSH tests above check these are called correctly.
    """

    @classmethod
    def setUpClass(cls):
        ssh.SSHClient = FakeSSHClient
        ssh.SFTPClient = FakeSFTPClient
        ssh.RSAKey = FakeRSAKey
        ssh.AgentRequestHandler = FakeAgentRequestHandler

    @classmethod
    def tearDownClass(cls):
        ssh.SSHClient = SSHClient
        ssh.SFTPClient = SFTPClient
        ssh.RSAKey = RSAKey
        ssh.AgentRequestHandler = AgentRequestHandler


class FakeAgentRequestHandler:
    def __init__(self, arg):
        pass


class FakeChannel:
    def __init__(self, exit_status):
        self.exit_status = exit_status

    def exit_status_ready(self):
        return True

    def recv_exit_status(self):
        return self.exit_status

    def write(self, data):
        pass

    def close(self):
        pass


class FakeBuffer:
    def __init__(self, data, channel):
        self.channel = channel
        self.data = data

    def __iter__(self):
        return iter(self.data)


class FakeSSHClient:
    def load_system_host_keys(self):
        pass

    def set_missing_host_key_policy(self, _):
        pass

    def connect(self, hostname, *args, **kwargs):
        if isclass(hostname) and issubclass(hostname, Exception):
            raise hostname()

    def get_transport(self):
        return self

    def open_session(self):
        pass

    def exec_command(self, command, get_pty=None):
        channel = FakeChannel(0)
        return (
            channel,
            FakeBuffer([], channel),
            FakeBuffer([], channel),
        )


class FakeSFTPClient:
    @classmethod
    def from_transport(cls, transport):
        return cls()

    def putfo(self, file_io, remote_location):
        pass

    def getfo(self, remote_location, file_io):
        pass


class FakeRSAKey:
    @classmethod
    def from_private_key_file(cls, *args, **kwargs):
        return cls()



import re

from pyinfra.api import FactBase

from .util.packaging import parse_packages

DEB_PACKAGE_NAME_REGEX = r"[a-zA-Z0-9\+\-\.]+"
DEB_PACKAGE_VERSION_REGEX = r"[a-zA-Z0-9:~\.\-\+]+"


class DebArch(FactBase):
    """
    Returns the architecture string used in apt repository sources, eg ``amd64``.
    """

    command = "dpkg --print-architecture"
    requires_command = "dpkg"


class DebPackages(FactBase):
    """
    Returns a dict of installed dpkg packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "dpkg -l"
    requires_command = "dpkg"

    default = dict

    regex = r"^[i|h]i\s+({0}):?[a-zA-Z0-9]*\s+({1}).+$".format(
        DEB_PACKAGE_NAME_REGEX,
        DEB_PACKAGE_VERSION_REGEX,
    )

    def process(self, output):
        return parse_packages(self.regex, output)


class DebPackage(FactBase):
    """
    Returns information on a .deb archive or installed package.
    """

    _regexes = {
        "name": r"^Package: ({0})$".format(DEB_PACKAGE_NAME_REGEX),
        "version": r"^Version: ({0})$".format(DEB_PACKAGE_VERSION_REGEX),
    }

    requires_command = "dpkg"

    def command(self, name):
        return "! test -e {0} && (dpkg -s {0} 2>/dev/null || true) || dpkg -I {0}".format(name)

    def process(self, output):
        data = {}

        for line in output:
            line = line.strip()
            for key, regex in self._regexes.items():
                matches = re.match(regex, line)
                if matches:
                    value = matches.group(1)
                    data[key] = value
                    break

        return data



from pyinfra.api import FactBase

# Mapping for iptables code arguments to variable names
IPTABLES_ARGS = {
    "-A": "chain",
    "-j": "jump",
    # Boolean matches
    "-p": "protocol",
    "-s": "source",
    "-d": "destination",
    "-i": "in_interface",
    "-o": "out_interface",
    # Logging
    "--log-prefix": "log_prefix",
    # NAT exit rules
    "--to-destination": "to_destination",
    "--to-source": "to_source",
    "--to-ports": "to_ports",
}


def parse_iptables_rule(line):
    """
    Parse one iptables rule. Returns a dict where each iptables code argument
    is mapped to a name using IPTABLES_ARGS.
    """

    bits = line.split()

    definition = {}

    key = None
    args = []
    not_arg = False

    def add_args():
        arg_string = " ".join(args)

        if key in IPTABLES_ARGS:
            definition_key = "not_{0}".format(IPTABLES_ARGS[key]) if not_arg else IPTABLES_ARGS[key]
            definition[definition_key] = arg_string
        else:
            definition.setdefault("extras", []).extend((key, arg_string))

    for bit in bits:
        if bit == "!":
            if key:
                add_args()
                args = []
                key = None

            not_arg = True

        elif bit.startswith("-"):
            if key:
                add_args()
                args = []
                not_arg = False

            key = bit

        else:
            args.append(bit)

    if key:
        add_args()

    if "extras" in definition:
        definition["extras"] = set(definition["extras"])

    return definition


class IptablesRules(FactBase):
    """
    Returns a list of iptables rules for a specific table:

    .. code:: python

        [
            {
                "chain": "PREROUTING",
                "jump": "DNAT",
            },
        ]
    """

    default = list

    def command(self, table="filter"):
        return "iptables-save -t {0}".format(table)

    def process(self, output):
        rules = []

        for line in output:
            if line.startswith("-"):
                rules.append(parse_iptables_rule(line))

        return rules


class Ip6tablesRules(IptablesRules):
    """
    Returns a list of ip6tables rules for a specific table:

    .. code:: python

        [
            {
                "chain": "PREROUTING",
                "jump": "DNAT",
            },
        ]
    """

    def command(self, table="filter"):
        return "ip6tables-save -t {0}".format(table)


class IptablesChains(FactBase):
    """
    Returns a dict of iptables chains & policies:

    .. code:: python

        {
            "NAME": "POLICY",
        }
    """

    default = dict

    def command(self, table="filter"):
        return "iptables-save -t {0}".format(table)

    def process(self, output):
        chains = {}

        for line in output:
            if line.startswith(":"):
                line = line[1:]

                name, policy, _ = line.split()
                chains[name] = policy

        return chains


class Ip6tablesChains(IptablesChains):
    """
    Returns a dict of ip6tables chains & policies:

    .. code:: python

        {
            "NAME": "POLICY",
        }
    """

    def command(self, table="filter"):
        return "ip6tables-save -t {0}".format(table)



from os import path
from unittest import TestCase

from pyinfra.context import ctx_state
from pyinfra_cli.main import _main

from ..paramiko_util import PatchSSHTestCase
from .util import run_cli


class TestCliEagerFlags(TestCase):
    def test_print_help(self):
        result = run_cli("--version")
        assert result.exit_code == 0, result.stdout

        result = run_cli("--help")
        assert result.exit_code == 0, result.stdout


class TestDeployCli(PatchSSHTestCase):
    def setUp(self):
        ctx_state.reset()

    def test_invalid_deploy(self):
        result = run_cli(
            "@local",
            "not-a-file.py",
        )
        assert result.exit_code == 1, result.stdout
        assert "No deploy file: not-a-file.py" in result.stdout


class TestOperationCli(PatchSSHTestCase):
    def test_invalid_operation_module(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "not_a_module.shell",
        )
        assert result.exit_code == 1, result.stdout
        assert "No such module: not_a_module"

    def test_invalid_operation_function(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "server.not_an_operation",
        )
        assert result.exit_code == 1, result.stdout
        assert "No such operation: server.not_an_operation"

    def test_deploy_operation(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "server.shell",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_deploy_operation_with_all(self):
        result = run_cli(
            path.join("tests", "deploy", "inventory_all.py"),
            "server.shell",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_deploy_operation_json_args(self):
        result = run_cli(
            path.join("tests", "deploy", "inventory_all.py"),
            "server.shell",
            '[["echo hi"], {}]',
        )
        assert result.exit_code == 0, result.stdout


class TestFactCli(PatchSSHTestCase):
    def test_get_fact(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "fact",
            "server.Os",
        )
        assert result.exit_code == 0, result.stdout
        assert '"somehost": null' in result.stdout

    def test_get_fact_with_kwargs(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "fact",
            "files.File",
            "path=.",
        )
        assert result.exit_code == 0, result.stdout
        assert '"somehost": null' in result.stdout

    def test_invalid_fact_module(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "fact",
            "not_a_module.NotAFact",
        )
        assert result.exit_code == 1, result.stdout
        assert "No such module: pyinfra.facts.not_a_module" in result.stdout

    def test_invalid_fact_class(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "fact",
            "server.NotAFact",
        )
        assert result.exit_code == 1, result.stdout
        assert "No such attribute in module pyinfra.facts.server: NotAFact" in result.stdout


class TestExecCli(PatchSSHTestCase):
    def test_exec_command(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_options(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--sudo",
            "--sudo-user",
            "pyinfra",
            "--su-user",
            "pyinfrawhat",
            "--port",
            "1022",
            "--user",
            "ubuntu",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_serial(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--serial",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_no_wait(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--no-wait",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_debug_operations(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--debug-operations",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout

    def test_exec_command_with_debug_facts(self):
        result = run_cli(
            path.join("tests", "deploy", "inventories", "inventory.py"),
            "exec",
            "--debug-facts",
            "--",
            "echo hi",
        )
        assert result.exit_code == 0, result.stdout


class TestDirectMainExecution(PatchSSHTestCase):
    """
    These tests are very similar as above, without the click wrappers - basically
    here because coverage.py fails to properly detect all the code under the wrapper.
    """

    def test_deploy_operation_direct(self):
        with self.assertRaises(SystemExit) as e:
            _main(
                inventory=path.join("tests", "test_deploy", "inventories", "inventory.py"),
                operations=["server.shell", "echo hi"],
                chdir=None,
                group_data=None,
                verbosity=0,
                ssh_user=None,
                ssh_port=None,
                ssh_key=None,
                ssh_key_password=None,
                ssh_password=None,
                sudo=False,
                sudo_user=None,
                use_sudo_password=False,
                su_user=None,
                parallel=None,
                fail_percent=0,
                dry=False,
                limit=None,
                no_wait=False,
                serial=False,
                winrm_username=None,
                winrm_password=None,
                winrm_port=None,
                winrm_transport=None,
                shell_executable=None,
                quiet=False,
                data=tuple(),
                debug=False,
                debug_facts=False,
                debug_operations=False,
                config_filename="config.py",
            )
            assert e.args == (0,)



from pyinfra.api import FactBase

from .util.packaging import parse_packages


class XbpsPackages(FactBase):
    """
    Returns a dict of installed XBPS packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "xbps-query -l"
    requires_command = "xbps-query"

    default = dict

    regex = r"^.. ([a-zA-Z0-9_\-\+]+)\-([0-9a-z_\.]+)"

    def process(self, output):
        return parse_packages(self.regex, output)



from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Optional, Union

import click
from gevent.lock import BoundedSemaphore

from pyinfra import logger
from pyinfra.connectors.util import remove_any_sudo_askpass_file

from .connectors import get_execution_connector
from .exceptions import ConnectError
from .facts import create_host_fact, delete_host_fact, get_host_fact, reload_host_fact

if TYPE_CHECKING:
    from pyinfra.api.inventory import Inventory
    from pyinfra.api.state import State


def extract_callable_datas(datas: List[Union[Callable[..., Any], Any]]) -> Generator[Any, Any, Any]:
    for data in datas:
        # Support for dynamic data, ie @deploy wrapped data defaults where
        # the data is stored on the state temporarily.
        if callable(data):
            data = data()

        yield data


class HostData:
    """
    Combines multiple AttrData's to search for attributes.
    """

    override_datas: Dict[str, Any]

    def __init__(self, host: "Host", *datas):
        self.__dict__["host"] = host

        parsed_datas = list(datas)

        # Inject an empty override data so we can assign during deploy
        self.__dict__["override_datas"] = {}
        parsed_datas.insert(0, self.override_datas)

        self.__dict__["datas"] = tuple(parsed_datas)

    def __getattr__(self, key: str):
        for data in extract_callable_datas(self.datas):
            try:
                return data[key]
            except KeyError:
                pass

        raise AttributeError(f"Host `{self.host}` has no data `{key}`")

    def __setattr__(self, key: str, value: Any):
        self.override_datas[key] = value

    def __str__(self):
        return str(self.datas)

    def get(self, key: str, default=None):
        return getattr(self, key, default)

    def dict(self):
        out = {}

        # Copy and reverse data objects (such that the first items override
        # the last, matching __getattr__ output).
        datas = list(self.datas)
        datas.reverse()

        for data in extract_callable_datas(datas):
            out.update(data)

        return out


class Host:
    """
    Represents a target host. Thin class that links up to facts and host/group
    data.
    """

    connection = None
    state: "State"

    # Current context inside an @operation function (op gen stage)
    in_op: bool = False
    current_op_hash: Optional[str] = None
    current_op_global_kwargs: Dict[str, Any]

    # Current context inside a @deploy function (op gen stage)
    in_deploy: bool = False
    current_deploy_name: Optional[str] = None
    current_deploy_kwargs = None
    current_deploy_data = None

    # Current context during operation execution
    executing_op_hash = None
    nested_executing_op_hash = None

    loop_position: List[int]

    # Arbitrary data dictionary connectors may use
    connector_data: Dict[str, Any]

    def loop(self, iterable):
        self.loop_position.append(0)
        for i, item in enumerate(iterable):
            self.loop_position[-1] = i
            yield item
        self.loop_position.pop()

    def __init__(
        self,
        name: str,
        inventory: "Inventory",
        groups,
        executor=get_execution_connector("ssh"),
    ):
        self.inventory = inventory
        self.groups = groups
        self.executor = executor
        self.name = name

        self.loop_position = []

        self.connector_data = {}
        self.current_op_global_kwargs = {}

        # Fact data store
        # TODO: how to not have Any here?
        self.facts: Dict[str, Any] = {}
        self.facts_lock = BoundedSemaphore()

        # Append only list of operation hashes as called on this host, used to
        # generate a DAG to create the final operation order.
        self.op_hash_order: List[str] = []

        # Create the (waterfall data: override, host, group, global)
        self.data = HostData(
            self,
            lambda: inventory.get_override_data(),
            lambda: inventory.get_host_data(name),
            lambda: inventory.get_groups_data(groups),
            lambda: inventory.get_data(),
            # @deploy function data are default values, so come last
            self.get_deploy_data,
        )

    def __str__(self):
        return "{0}".format(self.name)

    def __repr__(self):
        return "Host({0})".format(self.name)

    @property
    def connected(self) -> bool:
        return self.connection is not None

    @property
    def host_data(self):
        return self.inventory.get_host_data(self.name)

    @property
    def group_data(self):
        return self.inventory.get_groups_data(self.groups)

    @property
    def print_prefix(self):
        if self.nested_executing_op_hash:
            return "{0}[{1}] {2} ".format(
                click.style(""),  # reset
                click.style(self.name, bold=True),
                click.style("nested", "blue"),
            )

        return "{0}[{1}] ".format(
            click.style(""),  # reset
            click.style(self.name, bold=True),
        )

    def style_print_prefix(self, *args, **kwargs):
        return "{0}[{1}] ".format(
            click.style(""),  # reset
            click.style(self.name, *args, **kwargs),
        )

    def get_deploy_data(self):
        if self.current_deploy_data:
            return self.current_deploy_data

        return {}

    def noop(self, description):
        """
        Log a description for a noop operation.
        """

        handler = logger.info if self.state.print_noop_info else logger.debug
        handler("{0}noop: {1}".format(self.print_prefix, description))

    @contextmanager
    def deploy(self, name: str, kwargs, data, in_deploy: bool = True):
        """
        Wraps a group of operations as a deploy, this should not be used
        directly, instead use ``pyinfra.api.deploy.deploy``.
        """

        # Handle nested deploy names
        if self.current_deploy_name:
            name = "{0} | {1}".format(self.current_deploy_name, name)

        # Store the previous values
        old_in_deploy = self.in_deploy
        old_deploy_name = self.current_deploy_name
        old_deploy_kwargs = self.current_deploy_kwargs
        old_deploy_data = self.current_deploy_data
        self.in_deploy = in_deploy

        # Set the new values
        self.current_deploy_name = name
        self.current_deploy_kwargs = kwargs
        self.current_deploy_data = data
        logger.debug(
            "Starting deploy %s (args=%r, data=%r)",
            name,
            kwargs,
            data,
        )

        yield

        # Restore the previous values
        self.in_deploy = old_in_deploy
        self.current_deploy_name = old_deploy_name
        self.current_deploy_kwargs = old_deploy_kwargs
        self.current_deploy_data = old_deploy_data

        logger.debug(
            "Reset deploy to %s (args=%r, data=%r)",
            old_deploy_name,
            old_deploy_kwargs,
            old_deploy_data,
        )

    # Host facts
    #

    def get_fact(self, name_or_cls, *args, **kwargs):
        """
        Get a fact for this host, reading from the cache if present.
        """
        return get_host_fact(self.state, self, name_or_cls, args=args, kwargs=kwargs)

    def reload_fact(self, name_or_cls, *args, **kwargs):
        """
        Get a fact for this host without using any cached value, always re-fetch the fact data
        from the host and then cache it.
        """
        return reload_host_fact(self.state, self, name_or_cls, args=args, kwargs=kwargs)

    def create_fact(self, name_or_cls, data=None, kwargs=None):
        """
        Create a new fact for this host in the fact cache.
        """
        return create_host_fact(self.state, self, name_or_cls, data, kwargs=kwargs)

    def delete_fact(self, name_or_cls, kwargs=None):
        """
        Remove an existing fact for this host in the fact cache.
        """
        return delete_host_fact(self.state, self, name_or_cls, kwargs=kwargs)

    # Connector proxy
    #

    def _check_state(self):
        if not self.state:
            raise TypeError("Cannot call this function with no state!")

    def connect(self, reason=None, show_errors: bool = True, raise_exceptions: bool = False):
        """
        Connect to the host using it's configured connector.
        """

        self._check_state()
        if not self.connection:
            self.state.trigger_callbacks("host_before_connect", self)

            try:
                self.connection = self.executor.connect(self.state, self)
            except ConnectError as e:
                if show_errors:
                    log_message = "{0}{1}".format(
                        self.print_prefix,
                        click.style(e.args[0], "red"),
                    )
                    logger.error(log_message)

                self.state.trigger_callbacks("host_connect_error", self, e)

                if raise_exceptions:
                    raise
            else:
                log_message = "{0}{1}".format(
                    self.print_prefix,
                    click.style("Connected", "green"),
                )
                if reason:
                    log_message = "{0}{1}".format(
                        log_message,
                        " ({0})".format(reason),
                    )

                logger.info(log_message)
                self.state.trigger_callbacks("host_connect", self)

        return self.connection

    def disconnect(self):
        """
        Disconnect from the host using it's configured connector.
        """
        self._check_state()

        # Disconnect is an optional function for executors if needed
        disconnect_func = getattr(self.executor, "disconnect", None)
        if disconnect_func:
            return disconnect_func(self.state, self)

        # TODO: consider whether this should be here!
        remove_any_sudo_askpass_file(self)

        self.state.trigger_callbacks("host_disconnect", self)

    def run_shell_command(self, *args, **kwargs):
        """
        Low level method to execute a shell command on the host via it's configured connector.
        """
        self._check_state()
        return self.executor.run_shell_command(self.state, self, *args, **kwargs)

    def put_file(self, *args, **kwargs):
        """
        Low level method to upload a file to the host via it's configured connector.
        """
        self._check_state()
        return self.executor.put_file(self.state, self, *args, **kwargs)

    def get_file(self, *args, **kwargs):
        """
        Low level method to download a file from the host via it's configured connector.
        """
        self._check_state()
        return self.executor.get_file(self.state, self, *args, **kwargs)

    # Rsync - optional executor specific ability

    def check_can_rsync(self):
        check_can_rsync_func = getattr(self.executor, "check_can_rsync", None)
        if check_can_rsync_func:
            return check_can_rsync_func(self)

        raise NotImplementedError(
            "The {0} connector does not support rsync!".format(
                self.executor.__name__,
            ),
        )

    def rsync(self, *args, **kwargs):
        self._check_state()
        return self.executor.rsync(self.state, self, *args, **kwargs)



"""
Manage pip (python) packages. Compatible globally or inside
a virtualenv (virtual environment).
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.files import File
from pyinfra.facts.pip import PipPackages

from . import files
from .util.packaging import ensure_packages


@operation
def virtualenv(
    path,
    python=None,
    venv=False,
    site_packages=False,
    always_copy=False,
    present=True,
):
    """
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + venv: use standard venv module instead of virtualenv
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    **Example:**

    .. code:: python

        pip.virtualenv(
            name="Create a virtualenv",
            path="/usr/local/bin/venv",
        )
    """

    # Check for *contents* of a virtualenv, ie don't accept an empty directory
    # as a valid virtualenv but ensure the activate script exists.
    activate_script_path = "{0}/bin/activate".format(path)

    if present is False:
        if host.get_fact(File, path=activate_script_path):
            yield from files.directory(path, present=False)
        else:
            host.noop("virtualenv {0} does not exist".format(path))

    if present:
        if not host.get_fact(File, path=activate_script_path):
            # Create missing virtualenv
            command = ["virtualenv"]

            if venv:
                command = [python or "python", "-m", "venv"]

            if python and not venv:
                command.append("-p {0}".format(python))

            if site_packages:
                command.append("--system-site-packages")

            if always_copy and not venv:
                command.append("--always-copy")
            elif always_copy and venv:
                command.append("--copies")

            command.append(path)

            yield " ".join(command)

            host.create_fact(
                File,
                kwargs={"path": activate_script_path},
                data={"user": None, "group": None},
            )
        else:
            host.noop("virtualenv {0} exists".format(path))


_virtualenv = virtualenv  # noqa


@operation
def venv(
    path,
    python=None,
    site_packages=False,
    always_copy=False,
    present=True,
):
    """
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    **Example:**

    .. code:: python

        pip.venv(
            name="Create a virtualenv",
            path="/usr/local/bin/venv",
        )
    """

    yield from virtualenv(
        venv=True,
        path=path,
        python=python,
        site_packages=site_packages,
        always_copy=always_copy,
        present=present,
    )


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    requirements=None,
    pip="pip",
    virtualenv=None,
    virtualenv_kwargs=None,
    extra_install_args=None,
):
    """
    Install/remove/update pip packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + requirements: location of requirements file to install/uninstall
    + pip: name or path of the pip directory to use
    + virtualenv: root directory of virtualenv to work in
    + virtualenv_kwargs: dictionary of arguments to pass to ``pip.virtualenv``
    + extra_install_args: additional arguments to the pip install command

    Virtualenv:
        This will be created if it does not exist already. ``virtualenv_kwargs``
        will be passed to ``pip.virtualenv`` which can be used to control how
        the env is created.

    Versions:
        Package versions can be pinned like pip: ``<pkg>==<version>``.

    **Example:**

    .. code:: python

        pip.packages(
            name="Install pyinfra into a virtualenv",
            packages=["pyinfra"],
            virtualenv="/usr/local/bin/venv",
        )
    """

    virtualenv_kwargs = virtualenv_kwargs or {}

    # Ensure any virtualenv
    if virtualenv:
        yield from _virtualenv(virtualenv, **virtualenv_kwargs)

        # And update pip path
        virtualenv = virtualenv.rstrip("/")
        pip = "{0}/bin/{1}".format(virtualenv, pip)

    install_command = [pip, "install"]
    if extra_install_args:
        install_command.append(extra_install_args)
    install_command = " ".join(install_command)

    uninstall_command = " ".join([pip, "uninstall", "--yes"])

    # (un)Install requirements
    if requirements is not None:
        if present:
            yield "{0} -r {1}".format(install_command, requirements)
        else:
            yield "{0} -r {1}".format(uninstall_command, requirements)

    # Handle passed in packages
    if packages:
        current_packages = host.get_fact(PipPackages, pip=pip)

        yield from ensure_packages(
            host,
            packages,
            current_packages,
            present,
            install_command=install_command,
            uninstall_command=uninstall_command,
            upgrade_command="{0} --upgrade".format(install_command),
            version_join="==",
            latest=latest,
        )



from pyinfra.api import FactBase

from .util import make_cat_files_command
from .util.packaging import parse_zypper_repositories


class ZypperRepositories(FactBase):
    """
    Returns a list of installed zypper repositories:

    .. code:: python

        [
            {
                "name": "Main Repository",
                "enabled": "1",
                "autorefresh": "0",
                "baseurl": "http://download.opensuse.org/distribution/leap/$releasever/repo/oss/",
                "type": "rpm-md",
            },
        ]
    """

    command = make_cat_files_command(
        "/etc/zypp/repos.d/*.repo",
    )
    requires_command = "zypper"

    default = list

    def process(self, output):
        return parse_zypper_repositories(output)



import re

from pyinfra.api import FactBase


class UpstartStatus(FactBase):
    """
    Returns a dict of name -> status for upstart managed services.
    """

    command = "initctl list"
    requires_command = "initctl"

    regex = r"^([a-z\-]+) [a-z]+\/([a-z]+)"
    default = dict

    def process(self, output):
        services = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                services[matches.group(1)] = matches.group(2) == "running"

        return services



import re

from pyinfra.api import FactBase


class OpenrcStatus(FactBase):
    """
    Returns a dict of name -> status for OpenRC services for a given runlevel.
    """

    default = dict
    requires_command = "rc-status"
    regex = (
        r"\s+([a-zA-Z0-9\-_]+)"
        r"\s+\[\s+"
        r"([a-z]+)"
        r"(?:\s(?:[0-9]+\sday\(s\)\s)?"
        r"[0-9]+\:[0-9]+\:[0-9]+\s\([0-9]+\))?"
        r"\s+\]"
    )

    def command(self, runlevel="default"):
        return "rc-status {0}".format(runlevel)

    def process(self, output):
        services = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                services[matches.group(1)] = matches.group(2) == "started"

        return services


class OpenrcEnabled(FactBase):
    """
    Returns a dict of name -> whether enabled for OpenRC services for a given runlevel.
    """

    default = dict
    requires_command = "rc-update"

    def command(self, runlevel="default"):
        self.runlevel = runlevel
        return "rc-update show -v"

    def process(self, output):
        services = {}

        for line in output:
            name, levels = line.split("|", 1)
            name = name.strip()
            levels = levels.split()
            services[name] = self.runlevel in levels

        return services



"""
The `ContextObject` and `ContextManager` provide context specific variables that
are imported and used throughout pyinfra and end user deploy code (CLI mode).

These variables always represent the current executing pyinfra context.
"""
from contextlib import contextmanager
from typing import TYPE_CHECKING

from gevent.local import local

if TYPE_CHECKING:
    from pyinfra.api.config import Config
    from pyinfra.api.host import Host
    from pyinfra.api.inventory import Inventory
    from pyinfra.api.state import State


class container:
    pass


class ContextObject:
    _container_cls = container
    _base_cls = None

    def __init__(self):
        self._container = self._container_cls()
        self._container.module = None

    def _get_module(self):
        return self._container.module

    def __repr__(self):
        return "ContextObject({0}):{1}".format(
            self._base_cls.__name__,
            repr(self._get_module()),
        )

    def __str__(self):
        return str(self._get_module())

    def __dir__(self):
        return dir(self._base_cls)

    def __getattr__(self, key):
        if self._get_module() is None:
            return getattr(self._base_cls, key)
        return getattr(self._get_module(), key)

    def __setattr__(self, key, value):
        if key in ("_container", "_base_cls"):
            return super().__setattr__(key, value)

        if self._get_module() is None:
            raise TypeError("Cannot assign to context base module")

        return setattr(self._get_module(), key, value)

    def __iter__(self):
        return iter(self._get_module())

    def __len__(self):
        return len(self._get_module())

    def __eq__(self, other):
        return self._get_module() == other

    def __hash__(self):
        return hash(self._get_module())


class LocalContextObject(ContextObject):
    _container_cls = local


class ContextManager:
    def __init__(self, key, context_cls):
        self.context = context_cls()

    def get(self):
        return getattr(self.context._container, "module", None)

    def set(self, module):
        self.context._container.module = module

    def set_base(self, module):
        self.context._base_cls = module

    def reset(self):
        self.context._container.module = None

    def isset(self):
        return self.get() is not None

    @contextmanager
    def use(self, module):
        old_module = self.get()
        self.set(module)
        yield
        self.set(old_module)


ctx_state = ContextManager("state", ContextObject)
state: "State" = ctx_state.context

ctx_inventory = ContextManager("inventory", ContextObject)
inventory: "Inventory" = ctx_inventory.context

# Config can be modified mid-deploy, so we use a local object here which
# is based on a copy of the state config.
ctx_config = ContextManager("config", LocalContextObject)
config: "Config" = ctx_config.context

# Hosts are prepared in parallel each in a greenlet, so we use a local to
# point at different host objects in each greenlet.
ctx_host = ContextManager("host", LocalContextObject)
host: "Host" = ctx_host.context


def init_base_classes() -> None:
    from pyinfra.api import Config, Host, Inventory, State

    ctx_config.set_base(Config)
    ctx_inventory.set_base(Inventory)
    ctx_host.set_base(Host)
    ctx_state.set_base(State)



from functools import wraps
from hashlib import sha1
from inspect import getframeinfo, getfullargspec, stack
from io import BytesIO, StringIO
from os import getcwd, path, stat
from socket import error as socket_error, timeout as timeout_error
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

import click
from jinja2 import Environment, FileSystemLoader, StrictUndefined
from paramiko import SSHException

import pyinfra
from pyinfra import logger

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State

# 64kb chunks
BLOCKSIZE = 65536

# Caches
TEMPLATES: Dict[Any, Any] = {}
FILE_SHAS: Dict[Any, Any] = {}

PYINFRA_API_DIR = path.dirname(__file__)


def get_file_path(state: "State", filename: str):
    if path.isabs(filename):
        return filename

    assert state.cwd is not None, "Cannot use `get_file_path` with no `state.cwd` set"
    relative_to = state.cwd

    if state.current_exec_filename and (filename.startswith("./") or filename.startswith(".\\")):
        relative_to = path.dirname(state.current_exec_filename)

    return path.join(relative_to, filename)


def get_args_kwargs_spec(func: Callable[..., Any]) -> Tuple[List, Dict]:
    args: List[Any] = []
    kwargs: Dict[Any, Any] = {}

    argspec = getfullargspec(func)
    if not argspec.args:
        return args, kwargs

    if argspec.defaults:
        kwargs = dict(
            zip(
                argspec.args[-len(argspec.defaults) :],
                argspec.defaults,
            ),
        )
        args = argspec.args[: -len(argspec.defaults)]
    else:
        args = argspec.args

    return args, kwargs


def get_kwargs_str(kwargs: Dict[Any, Any]):
    if not kwargs:
        return ""

    items = [
        "{0}={1}".format(key, value)
        for key, value in sorted(kwargs.items())
        if key not in ("self", "state", "host")
    ]
    return ", ".join(items)


def try_int(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        return value


def memoize(func: Callable[..., Any]):
    @wraps(func)
    def wrapper(*args, **kwargs):
        key = "{0}{1}".format(args, kwargs)
        if key in wrapper.cache:
            return wrapper.cache[key]

        value = func(*args, **kwargs)
        wrapper.cache[key] = value
        return value

    wrapper.cache = {}  # type: ignore
    return wrapper


def get_call_location(frame_offset: int = 1):
    frame = get_caller_frameinfo(frame_offset=frame_offset)  # escape *this* function
    relpath = frame.filename

    try:
        # On Windows if pyinfra is on a different drive to the filename here, this will
        # error as there's no way to do relative paths between drives.
        relpath = path.relpath(frame.filename)
    except ValueError:
        pass

    return "line {0} in {1}".format(frame.lineno, relpath)


def get_caller_frameinfo(frame_offset: int = 0):
    # Default frames to look at is 2; one for this function call itself
    # in util.py and one for the caller of this function within pyinfra
    # giving the external call frame (ie end user deploy code).
    frame_shift = 2 + frame_offset

    stack_items = stack()

    frame = stack_items[frame_shift][0]
    info = getframeinfo(frame)

    del stack_items

    return info


def get_operation_order_from_stack(state: "State"):
    stack_items = list(reversed(stack()))

    # Find the *first* occurrence of our deploy file in the reversed stack
    if state.current_deploy_filename:
        for i, stack_item in enumerate(stack_items):
            frame = getframeinfo(stack_item[0])
            if frame.filename == state.current_deploy_filename:
                break
    else:
        i = 0

    # Now generate a list of line numbers *following that file*
    line_numbers = []

    if pyinfra.is_cli:
        line_numbers.append(state.current_op_file_number)

    for stack_item in stack_items[i:]:
        frame = getframeinfo(stack_item[0])

        if frame.filename.startswith(PYINFRA_API_DIR):
            continue

        line_numbers.append(frame.lineno)

    del stack_items

    return line_numbers


def get_template(filename_or_io: str):
    """
    Gets a jinja2 ``Template`` object for the input filename or string, with caching
    based on the filename of the template, or the SHA1 of the input string.
    """

    file_data = get_file_io(filename_or_io, mode="r")
    cache_key = file_data.cache_key

    if cache_key and cache_key in TEMPLATES:
        return TEMPLATES[cache_key]

    with file_data as file_io:
        template_string = file_io.read()

    template = Environment(
        undefined=StrictUndefined,
        keep_trailing_newline=True,
        loader=FileSystemLoader(getcwd()),
    ).from_string(template_string)

    if cache_key:
        TEMPLATES[cache_key] = template

    return template


def sha1_hash(string: str):
    """
    Return the SHA1 of the input string.
    """

    hasher = sha1()
    hasher.update(string.encode("utf-8"))
    return hasher.hexdigest()


def format_exception(e):
    return "{0}{1}".format(e.__class__.__name__, e.args)


def print_host_combined_output(host: "Host", combined_output_lines):
    for type_, line in combined_output_lines:
        if type_ == "stderr":
            logger.error(
                "{0}{1}".format(
                    host.print_prefix,
                    click.style(line, "red"),
                ),
            )
        else:
            logger.error(
                "{0}{1}".format(
                    host.print_prefix,
                    line,
                ),
            )


def log_operation_start(op_meta: Dict, op_types: Optional[List] = None, prefix: str = "--> "):
    op_types = op_types or []
    if op_meta["serial"]:
        op_types.append("serial")
    if op_meta["run_once"]:
        op_types.append("run once")

    args = ""
    if op_meta["args"]:
        args = "({0})".format(", ".join(str(arg) for arg in op_meta["args"]))

    logger.info(
        "{0} {1} {2}".format(
            click.style(
                "{0}Starting{1}operation:".format(
                    prefix,
                    " {0} ".format(", ".join(op_types)) if op_types else " ",
                ),
                "blue",
            ),
            click.style(", ".join(op_meta["names"]), bold=True),
            args,
        ),
    )


def log_error_or_warning(
    host: "Host", ignore_errors: bool, description: str = "", continue_on_error: bool = False
):
    log_func = logger.error
    log_color = "red"
    log_text = "Error: " if description else "Error"

    if ignore_errors:
        log_func = logger.warning
        log_color = "yellow"
        log_text = (
            "Error (ignored, execution continued)" if continue_on_error else "Error (ignored)"
        )
        if description:
            log_text = f"{log_text}: "

    log_func(
        "{0}{1}{2}".format(
            host.print_prefix,
            click.style(log_text, log_color),
            description,
        ),
    )


def log_host_command_error(host: "Host", e, timeout: int = 0):
    if isinstance(e, timeout_error):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command timed out after {0}s".format(
                        timeout,
                    ),
                    "red",
                ),
            ),
        )

    elif isinstance(e, (socket_error, SSHException)):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command socket/SSH error: {0}".format(format_exception(e)),
                    "red",
                ),
            ),
        )

    elif isinstance(e, IOError):
        logger.error(
            "{0}{1}".format(
                host.print_prefix,
                click.style(
                    "Command IO error: {0}".format(format_exception(e)),
                    "red",
                ),
            ),
        )

    # Still here? Re-raise!
    else:
        raise e


def make_hash(obj):
    """
    Make a hash from an arbitrary nested dictionary, list, tuple or set, used to generate
    ID's for operations based on their name & arguments.
    """

    if isinstance(obj, (set, tuple, list)):
        hash_string = "".join([make_hash(e) for e in obj])

    elif isinstance(obj, dict):
        hash_string = "".join("".join((key, make_hash(value))) for key, value in obj.items())

    else:
        hash_string = (
            # Capture integers first (as 1 == True)
            "{0}".format(obj)
            if isinstance(obj, int)
            # Constants - the values can change between hosts but we should still
            # group them under the same operation hash.
            else "_PYINFRA_CONSTANT"
            if obj in (True, False, None)
            # Plain strings
            else obj
            if isinstance(obj, str)
            # Objects with __name__s
            else obj.__name__
            if hasattr(obj, "__name__")
            # Objects with names
            else obj.name
            if hasattr(obj, "name")
            # Repr anything else
            else repr(obj)
        )

    return sha1_hash(hash_string)


class get_file_io:
    """
    Given either a filename or an existing IO object, this context processor
    will open and close filenames, and leave IO objects alone.
    """

    close = False

    def __init__(self, filename_or_io, mode="rb"):
        if not (
            # Check we can be read
            hasattr(filename_or_io, "read")
            # Or we're a filename
            or isinstance(filename_or_io, str)
        ):
            raise TypeError(
                "Invalid filename or IO object: {0}".format(
                    filename_or_io,
                ),
            )

        # Convert any StringIO/BytesIO to the other to match the desired mode
        if isinstance(filename_or_io, StringIO) and mode == "rb":
            filename_or_io.seek(0)
            filename_or_io = BytesIO(filename_or_io.read().encode())
        if isinstance(filename_or_io, BytesIO) and mode == "r":
            filename_or_io.seek(0)
            filename_or_io = StringIO(filename_or_io.read().decode())

        self.filename_or_io = filename_or_io
        self.mode = mode

    def __enter__(self):
        # If we have a read attribute, just use the object as-is
        if hasattr(self.filename_or_io, "read"):
            file_io = self.filename_or_io

        # Otherwise, assume a filename and open it up
        else:
            file_io = open(self.filename_or_io, self.mode)

            # Attach to self for closing on __exit__
            self.file_io = file_io
            self.close = True

        # Ensure we're at the start of the file
        file_io.seek(0)
        return file_io

    def __exit__(self, type, value, traceback):
        if self.close:
            self.file_io.close()

    @property
    def cache_key(self):
        # If we're a filename, cache against that - we don't cache in-memory
        # file objects.
        if isinstance(self.filename_or_io, str):
            return self.filename_or_io


def get_file_sha1(filename_or_io):
    """
    Calculates the SHA1 of a file or file object using a buffer to handle larger files.
    """

    file_data = get_file_io(filename_or_io)
    cache_key = file_data.cache_key

    if cache_key and cache_key in FILE_SHAS:
        return FILE_SHAS[cache_key]

    with file_data as file_io:
        hasher = sha1()
        buff = file_io.read(BLOCKSIZE)

        while len(buff) > 0:
            if isinstance(buff, str):
                buff = buff.encode("utf-8")

            hasher.update(buff)
            buff = file_io.read(BLOCKSIZE)

    digest = hasher.hexdigest()

    if cache_key:
        FILE_SHAS[cache_key] = digest

    return digest


def get_path_permissions_mode(pathname: str):
    """
    Get the permissions (bits) of a path as an integer.
    """

    mode_octal = oct(stat(pathname).st_mode)
    return mode_octal[-3:]



def _remove_prefix(title, remove_prefix):
    if title.startswith(remove_prefix):
        return title[len(remove_prefix) :]

    return title


def parse_columns_and_rows(
    lines,
    delimiter,
    remove_column_prefix=None,
    title_parser=None,
):
    titles = lines[0]
    titles = titles.split(delimiter)

    if title_parser:
        titles = [title_parser(title) for title in titles]

    if remove_column_prefix:
        titles = [_remove_prefix(title, remove_column_prefix) for title in titles]

    rows = []

    for line in lines[1:]:
        bits = line.split(delimiter)
        row = {}

        for i, bit in enumerate(bits):
            row[titles[i]] = bit or None

        rows.append(row)

    return rows



"""
The files facts provide information about the filesystem and it's contents on the target host.
"""

import re
import stat
from datetime import datetime
from typing import List, Tuple

from pyinfra.api.command import QuoteString, make_formatted_string_command
from pyinfra.api.facts import FactBase
from pyinfra.api.util import try_int

LINUX_STAT_COMMAND = "stat -c 'user=%U group=%G mode=%A atime=%X mtime=%Y ctime=%Z size=%s %N'"
BSD_STAT_COMMAND = "stat -f 'user=%Su group=%Sg mode=%Sp atime=%a mtime=%m ctime=%c size=%z %N%SY'"

STAT_REGEX = (
    r"user=(.*) group=(.*) mode=(.*) "
    r"atime=([0-9]*) mtime=([0-9]*) ctime=([0-9]*) "
    r"size=([0-9]*) (.*)"
)

FLAG_TO_TYPE = {
    "b": "block",
    "c": "character",
    "d": "directory",
    "l": "link",
    "s": "socket",
    "p": "fifo",
    "-": "file",
}

# Each item is a map of character to permission octal to be combined, taken from stdlib:
# https://github.com/python/cpython/blob/c1c3be0f9dc414bfae9a5718451ca217751ac687/Lib/stat.py#L128-L154
CHAR_TO_PERMISSION = (
    # User
    {"r": stat.S_IRUSR},
    {"w": stat.S_IWUSR},
    {"x": stat.S_IXUSR, "S": stat.S_ISUID, "s": stat.S_IXUSR | stat.S_ISUID},
    # Group
    {"r": stat.S_IRGRP},
    {"w": stat.S_IWGRP},
    {"x": stat.S_IXGRP, "S": stat.S_ISGID, "s": stat.S_IXGRP | stat.S_ISGID},
    # Other
    {"r": stat.S_IROTH},
    {"w": stat.S_IWOTH},
    {"x": stat.S_IXOTH, "T": stat.S_ISVTX, "t": stat.S_IXOTH | stat.S_ISVTX},
)


def _parse_mode(mode: str) -> int:
    """
    Converts ls mode output (rwxrwxrwx) -> octal permission integer (755).
    """

    out = 0

    for i, char in enumerate(mode):
        for c, m in CHAR_TO_PERMISSION[i].items():
            if char == c:
                out |= m
                break

    return int(oct(out)[2:])


class File(FactBase):
    """
    Returns information about a file on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
            "size": 3928,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a file:
        returns ``False``
    """

    type = "file"

    def command(self, path):
        return make_formatted_string_command(
            (
                # only stat if the path exists (file or symlink)
                "! (test -e {0} || test -L {0} ) || "
                "( {linux_stat_command} {0} 2> /dev/null || {bsd_stat_command} {0} )"
            ),
            QuoteString(path),
            linux_stat_command=LINUX_STAT_COMMAND,
            bsd_stat_command=BSD_STAT_COMMAND,
        )

    def process(self, output):
        match = re.match(STAT_REGEX, output[0])
        if not match:
            return None

        data = {}
        path_type = None

        for key, value in (
            ("user", match.group(1)),
            ("group", match.group(2)),
            ("mode", match.group(3)),
            ("atime", match.group(4)),
            ("mtime", match.group(5)),
            ("ctime", match.group(6)),
            ("size", match.group(7)),
        ):
            if key == "mode":
                path_type = FLAG_TO_TYPE[value[0]]
                value = _parse_mode(value[1:])

            elif key == "size":
                value = try_int(value)

            elif key in ("atime", "mtime", "ctime"):
                value = try_int(value)
                if isinstance(value, int):
                    value = datetime.utcfromtimestamp(value)

            data[key] = value

        if path_type != self.type:
            return False

        if path_type == "link":
            filename = match.group(8)
            filename, target = filename.split(" -> ")
            data["link_target"] = target.strip("'").lstrip("`")

        return data


class Link(File):
    """
    Returns information about a link on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "link_target": "/path/to/link/target"
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a link:
        returns ``False``
    """

    type = "link"


class Directory(File):
    """
    Returns information about a directory on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
            "mode": 644,
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a directory:
        returns ``False``
    """

    type = "directory"


class Socket(File):
    """
    Returns information about a socket on the remote system:

    .. code:: python

        {
            "user": "pyinfra",
            "group": "pyinfra",
        }

    If the path does not exist:
        returns ``None``

    If the path exists but is not a socket:
        returns ``False``
    """

    type = "socket"


class HashFileFactBase(FactBase):
    _raw_cmd: str
    _regexes: Tuple[str, str]

    def __init_subclass__(cls, digits: int, cmds: List[str], **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        raw_hash_cmds = ["%s {0} 2> /dev/null" % cmd for cmd in cmds]
        raw_hash_cmd = " || ".join(raw_hash_cmds)
        cls._raw_cmd = "test -e {0} && ( %s ) || true" % raw_hash_cmd

        assert cls.__name__.endswith("File")
        hash_name = cls.__name__[:-4].upper()
        cls._regexes = (
            # GNU coreutils style:
            r"^([a-fA-F0-9]{%d})\s+%%s$" % digits,
            # BSD style:
            r"^%s\s+\(%%s\)\s+=\s+([a-fA-F0-9]{%d})$" % (hash_name, digits),
        )

    def command(self, path):
        self.path = path
        return make_formatted_string_command(self._raw_cmd, QuoteString(path))

    def process(self, output):
        output = output[0]
        escaped_path = re.escape(self.path)
        for regex in self._regexes:
            matches = re.match(regex % escaped_path, output)
            if matches:
                return matches.group(1)


class Sha1File(HashFileFactBase, digits=40, cmds=["sha1sum", "shasum", "sha1"]):
    """
    Returns a SHA1 hash of a file. Works with both sha1sum and sha1. Returns
    ``None`` if the file doest not exist.
    """


class Sha256File(HashFileFactBase, digits=64, cmds=["sha256sum", "shasum -a 256", "sha256"]):
    """
    Returns a SHA256 hash of a file, or ``None`` if the file does not exist.
    """


class Md5File(HashFileFactBase, digits=32, cmds=["md5sum", "md5"]):
    """
    Returns an MD5 hash of a file, or ``None`` if the file does not exist.
    """


class FindInFile(FactBase):
    """
    Checks for the existence of text in a file using grep. Returns a list of matching
    lines if the file exists, and ``None`` if the file does not.
    """

    def command(self, path, pattern, interpolate_variables=False):
        self.exists_flag = "__pyinfra_exists_{0}".format(path)

        if interpolate_variables:
            pattern = '"{0}"'.format(pattern.replace('"', '\\"'))
        else:
            pattern = QuoteString(pattern)

        return make_formatted_string_command(
            (
                "grep -e {0} {1} 2> /dev/null || "
                "( find {1} -type f > /dev/null && echo {2} || true )"
            ),
            pattern,
            QuoteString(path),
            QuoteString(self.exists_flag),
        )

    def process(self, output):
        # If output is the special string: no matches, so return an empty list;
        # this allows us to differentiate between no matches in an existing file
        # or a file not existing.
        if output and output[0] == self.exists_flag:
            return []

        return output


class FindFilesBase(FactBase):
    abstract = True
    default = list

    @staticmethod
    def process(output):
        return output

    def command(self, path, quote_path=True):
        return make_formatted_string_command(
            "find {0} -type {type_flag} || true",
            QuoteString(path) if quote_path else path,
            type_flag=self.type_flag,
        )


class FindFiles(FindFilesBase):
    """
    Returns a list of files from a start path, recursively using ``find``.
    """

    type_flag = "f"


class FindLinks(FindFilesBase):
    """
    Returns a list of links from a start path, recursively using ``find``.
    """

    type_flag = "l"


class FindDirectories(FindFilesBase):
    """
    Returns a list of directories from a start path, recursively using ``find``.
    """

    type_flag = "d"


class Flags(FactBase):
    """
    Returns a list of the file flags set for the specified file or directory.
    """

    requires_command = "chflags"  # don't try to retrieve them if we can't set them

    def command(self, path):
        return make_formatted_string_command(
            "! test -e {0} || stat -f %Sf {0}",
            QuoteString(path),
        )

    def process(self, output):

        return [flag for flag in output[0].split(",") if len(flag) > 0] if len(output) == 1 else []


MARKER_DEFAULT = "# {mark} PYINFRA BLOCK"
MARKER_BEGIN_DEFAULT = "BEGIN"
MARKER_END_DEFAULT = "END"
EXISTS = "__pyinfra_exists_"
MISSING = "__pyinfra_missing_"


class Block(FactBase):
    """
    Returns a (possibly empty) list of the lines found between the markers.

    .. code:: python

        [
            "xray: one",
            "alpha: two"
        ]

    If the ``path`` doesn't exist
        returns ``None``

    If the ``path`` exists but the markers are not found
        returns ``[]``
    """

    # if markers aren't found, awk will return 0 and produce no output but we need to
    # distinguish between "markers not found" and "markers found but nothing between them"
    # for the former we use the empty list (created the call to default) and for the latter
    # the list with a single empty string.
    default = list

    def command(self, path, marker=None, begin=None, end=None):

        self.path = path
        start = (marker or MARKER_DEFAULT).format(mark=begin or MARKER_BEGIN_DEFAULT)
        end = (marker or MARKER_DEFAULT).format(mark=end or MARKER_END_DEFAULT)
        if start == end:
            raise ValueError(f"delimiters for block must be different but found only '{start}'")

        backstop = make_formatted_string_command(
            "(find {0} -type f > /dev/null && echo {1} || echo {2} )",
            QuoteString(path),
            QuoteString(f"{EXISTS}{path}"),
            QuoteString(f"{MISSING}{path}"),
        )
        # m_f_s_c inserts blanks in unfortunate places, e.g. after first slash
        cmd = make_formatted_string_command(
            f"awk \\'/{end}/{{{{f=0}}}} f; /{start}/{{{{f=1}}}}\\' {{0}} || {backstop}",
            QuoteString(path),
        )
        return cmd

    def process(self, output):
        if output and (output[0] == f"{EXISTS}{self.path}"):
            return []
        if output and (output[0] == f"{MISSING}{self.path}"):
            return None
        return output



"""
This file as originally part of the "sshuserclient" pypi package. The GitHub
source has now vanished (https://github.com/tobald/sshuserclient).
"""

import glob
import re
from os import environ, path

import paramiko.config
from gevent.subprocess import CalledProcessError, check_call
from paramiko import SSHConfig as ParamikoSSHConfig

from pyinfra import logger

SETTINGS_REGEX = re.compile(r"(\w+)(?:\s*=\s*|\s+)(.+)")


class FakeInvokeResult:
    ok = False


class FakeInvoke:
    @staticmethod
    def run(cmd, *args, **kwargs):
        result = FakeInvokeResult()

        try:
            cmd = [environ["SHELL"], cmd]
            try:
                code = check_call(cmd)
            except CalledProcessError as e:
                code = e.returncode
            result.ok = code == 0
        except Exception as e:
            logger.warning(
                ("pyinfra encountered an error loading SSH config match exec {0}: {1}").format(
                    cmd,
                    e,
                ),
            )

        return result


paramiko.config.invoke = FakeInvoke  # type: ignore


def _expand_include_statements(file_obj, parsed_files=None):
    parsed_lines = []

    for line in file_obj.readlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        match = re.match(SETTINGS_REGEX, line)
        if not match:
            parsed_lines.append(line)
            continue

        key = match.group(1).lower()
        value = match.group(2)

        if key != "include":
            parsed_lines.append(line)
            continue

        if parsed_files is None:
            parsed_files = []

        # The path can be relative to its parent configuration file
        if path.isabs(value) is False and value[0] != "~":
            folder = path.dirname(file_obj.name)
            value = path.join(folder, value)

        value = path.expanduser(value)

        for filename in glob.iglob(value):
            if path.isfile(filename):
                if filename in parsed_files:
                    raise Exception(
                        "Include loop detected in ssh config file: %s" % filename,
                    )
                with open(filename, encoding="utf-8") as fd:
                    parsed_files.append(filename)
                    parsed_lines.extend(_expand_include_statements(fd, parsed_files))

    return parsed_lines


class SSHConfig(ParamikoSSHConfig):
    """
    an SSHConfig that supports includes directives
    https://github.com/paramiko/paramiko/pull/1194
    """

    def parse(self, file_obj):
        file_obj = _expand_include_statements(file_obj)
        return super().parse(file_obj)



from typing import (
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    Protocol,
    Tuple,
    overload,
)

from typing_extensions import ParamSpec

from pyinfra.api.command import (
    FileDownloadCommand,
    FileUploadCommand,
    FunctionCommand,
    StringCommand,
)
from pyinfra.api.host import Host
from pyinfra.api.state import State

P = ParamSpec("P")

Command = str | StringCommand | FileDownloadCommand | FileUploadCommand | FunctionCommand

class OperationMeta:
    changed: bool
    commands: List[str] | None
    hash: str | None

    stdout_lines: List[str]
    stdout: str
    stderr_lines: List[str]
    stderr: str

class Operation(Generic[P], Protocol):
    def __call__(
        self,
        _sudo: bool | None = None,
        _sudo_user: str | None = None,
        _use_sudo_login: bool | None = None,
        _use_sudo_password: bool | None = None,
        _preserve_sudo_env: bool | None = None,
        _su_user: str | None = None,
        _use_su_login: bool | None = None,
        _preserve_su_env: bool | None = None,
        _su_shell: str | None = None,
        _doas: bool | None = None,
        _doas_user: str | None = None,
        _shell_executable: str | None = None,
        _chdir: str | None = None,
        _env: Mapping[str, str] | None = None,
        _success_exit_codes: Iterable[int] | None = None,
        _timeout: int | None = None,
        _get_pty: bool | None = None,
        _stdin: str | List[str] | Tuple[str, ...] | None = None,
        name: str | None = None,
        _ignore_errors: bool | None = None,
        _continue_on_error: bool | None = None,
        _precondition: str | None = None,
        _postcondition: str | None = None,
        _on_success: Callable[[State, Host, str], None] | None = None,
        _on_error: Callable[[State, Host, str], None] | None = None,
        _parallel: int | None = None,
        _run_once: bool | None = None,
        _serial: bool | None = None,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> OperationMeta: ...

def add_op(
    state: State,
    op_func: Operation[P],
    _sudo: bool | None = None,
    _sudo_user: str | None = None,
    _use_sudo_login: bool | None = None,
    _use_sudo_password: bool | None = None,
    _preserve_sudo_env: bool | None = None,
    _su_user: str | None = None,
    _use_su_login: bool | None = None,
    _preserve_su_env: bool | None = None,
    _su_shell: str | None = None,
    _doas: bool | None = None,
    _doas_user: str | None = None,
    _shell_executable: str | None = None,
    _chdir: str | None = None,
    _env: Mapping[str, str] | None = None,
    _success_exit_codes: Iterable[int] | None = None,
    _timeout: int | None = None,
    _get_pty: bool | None = None,
    _stdin: str | List[str] | Tuple[str, ...] | None = None,
    name: str | None = None,
    _ignore_errors: bool | None = None,
    _continue_on_error: bool | None = None,
    _precondition: str | None = None,
    _postcondition: str | None = None,
    _on_success: Callable[[State, Host, str], None] | None = None,
    _on_error: Callable[[State, Host, str], None] | None = None,
    _parallel: int | None = None,
    _run_once: bool | None = None,
    _serial: bool | None = None,
    host: Iterable[Host] | Host | None = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> Dict[Host, OperationMeta]: ...
@overload
def operation(func: Callable[P, Generator[Command, None, None]]) -> Operation[P]: ...
@overload
def operation(
    pipeline_facts=None,
    is_idempotent: bool = True,
    idempotent_notice=None,
    frame_offset=1,
) -> Callable[[Callable[P, Generator[Command, None, None]]], Operation[P]]: ...



import json

from .exceptions import CliError
from .util import parse_cli_arg, try_import_module_attribute


def get_func_and_args(commands):
    operation_name = commands[0]

    op = try_import_module_attribute(operation_name, prefix="pyinfra.operations")

    # Parse the arguments
    operation_args = commands[1:]

    if len(operation_args) == 1:
        # Check if we're JSON (in which case we expect a list of two items:
        # a list of args and a dict of kwargs).
        try:
            args, kwargs = json.loads(operation_args[0])
            return op, (args or (), kwargs or {})
        except ValueError:
            pass

    args = [parse_cli_arg(arg) for arg in operation_args if "=" not in arg]

    kwargs = {
        key: parse_cli_arg(value)
        for key, value in [arg.split("=", 1) for arg in operation_args if "=" in arg]
    }

    return op, (args, kwargs)


def get_facts_and_args(commands):
    facts = []

    current_fact = None

    for command in commands:
        if "=" in command:
            if not current_fact:
                raise CliError("Invalid fact commands: `{0}`".format(commands))

            key, value = command.split("=", 1)
            current_fact[2][key] = value
            continue

        if current_fact:
            facts.append(current_fact)
            current_fact = None

        if "." not in command:
            raise CliError(f"Invalid fact: `{command}`, should be in the format `module.cls`")

        fact_cls = try_import_module_attribute(command, prefix="pyinfra.facts")
        current_fact = (fact_cls, (), {})

    if current_fact:
        facts.append(current_fact)

    return facts



import json
import platform
import warnings
from importlib import import_module
from os import listdir, path
from unittest import TestCase
from unittest.mock import MagicMock, patch

from pyinfra.api import FileDownloadCommand, FileUploadCommand, FunctionCommand, StringCommand
from pyinfra.context import ctx_host, ctx_state
from pyinfra_cli.util import json_encode

from .util import FakeState, JsonTest, create_host, get_command_string, parse_value, patch_files

PLATFORM_NAME = platform.system()


def parse_commands(commands):
    json_commands = []

    for command in commands:
        if isinstance(command, str):  # matches pyinfra/api/operation.py
            command = StringCommand(command.strip())

        if isinstance(command, StringCommand):
            json_command = get_command_string(command)

        elif isinstance(command, dict):
            command["command"] = get_command_string(command["command"]).strip()
            json_command = command

        elif isinstance(command, FunctionCommand):
            func_name = (
                command.function if command.function == "__func__" else command.function.__name__
            )
            json_command = [
                func_name,
                list(command.args),
                command.kwargs,
            ]

        elif isinstance(command, FileUploadCommand):
            if hasattr(command.src, "read"):
                command.src.seek(0)
                data = command.src.read()
            else:
                data = str(command.src)
            json_command = ["upload", data, str(command.dest)]

        elif isinstance(command, FileDownloadCommand):
            json_command = ["download", str(command.src), str(command.dest)]

        else:
            raise Exception("{0} is not a valid command!".format(command))

        if command.executor_kwargs:
            command.executor_kwargs["command"] = json_command
            json_command = command.executor_kwargs

        json_commands.append(json_command)
    return json_commands


def assert_commands(commands, wanted_commands):
    try:
        assert commands == wanted_commands
    except AssertionError as e:
        print()
        print("--> COMMANDS OUTPUT:")
        print(json.dumps(commands, indent=4, default=json_encode))

        print("--> TEST WANTS:")
        print(
            json.dumps(
                wanted_commands,
                indent=4,
                default=json_encode,
            ),
        )

        raise e


def make_operation_tests(arg):
    # Get the operation we're testing against
    module_name, op_name = arg.split(".")
    module = import_module("pyinfra.operations.{0}".format(module_name))
    op = getattr(module, op_name)

    # Generate a test class
    @patch("pyinfra.operations.files.get_timestamp", lambda: "a-timestamp")
    @patch("pyinfra.operations.util.files.get_timestamp", lambda: "a-timestamp")
    @patch("pyinfra.operations.python.getfullargspec", lambda x: MagicMock())
    class TestTests(TestCase, metaclass=JsonTest):
        jsontest_files = path.join("tests", "operations", arg)
        jsontest_prefix = "test_{0}_{1}_".format(module_name, op_name)

        @classmethod
        def setUpClass(cls):
            # Create a global fake state that attach to context state
            cls.state = FakeState()

        def jsontest_function(self, test_name, test_data):
            if (
                "require_platform" in test_data
                and PLATFORM_NAME not in test_data["require_platform"]
            ):
                return

            op_test_name = "{0}/{1}.json".format(arg, test_name)

            # Create a host with this tests facts and attach to context host
            host = create_host(facts=test_data.get("facts", {}))

            allowed_exception = test_data.get("exception")

            args = parse_value(test_data.get("args", []))
            kwargs = parse_value(test_data.get("kwargs", {}))

            with ctx_state.use(self.state):
                with ctx_host.use(host):
                    with patch_files(test_data.get("local_files", {})):
                        try:
                            output_commands = list(op._pyinfra_op(*args, **kwargs))
                        except Exception as e:
                            if allowed_exception:
                                allowed_exception_names = allowed_exception.get("names")
                                if not allowed_exception_names:
                                    allowed_exception_names = [allowed_exception["name"]]

                                if e.__class__.__name__ not in allowed_exception_names:
                                    print("Wrong exception raised!")
                                    raise

                                assert e.args[0] == allowed_exception["message"]
                                return

                            raise

                        op_is_idempotent = getattr(op._pyinfra_op, "is_idempotent", True)
                        second_output_commands = []
                        test_second_output_commands = "second_output_commands" in test_data

                        if op_is_idempotent or test_second_output_commands:
                            second_output_commands = list(op._pyinfra_op(*args, **kwargs))

            if op_is_idempotent:
                if test_data.get("idempotent", True):
                    if second_output_commands:
                        raise Exception(
                            "Operation not idempotent, second output commands: {0}".format(
                                second_output_commands,
                            ),
                        )
                else:
                    if not second_output_commands:
                        raise Exception(
                            (
                                "Operation tests as idempotent but test " "says it is not: {0}"
                            ).format(op_test_name),
                        )
                    if not test_data.get("disable_idempotent_warning_reason"):
                        warnings.warn(
                            (
                                "This operation should be idempotent, but the test has "
                                "disabled this check without reason: {0}"
                            ).format(op_test_name),
                        )

            if op_is_idempotent is False:
                if "disable_idempotent_warning_reason" in test_data:
                    warnings.warn(
                        (
                            "This operation is not idempotent and so the test does not need "
                            "`disable_idempotent_warning_reason` set: {0}"
                        ).format(op_test_name),
                    )

            commands = parse_commands(output_commands)
            assert_commands(commands, test_data["commands"])

            if test_second_output_commands:
                assert_commands(
                    parse_commands(second_output_commands),
                    test_data["second_output_commands"],
                )

            noop_description = test_data.get("noop_description")
            if len(commands) == 0 or noop_description:
                if noop_description is not None:
                    assert host.noop_description == noop_description
                else:
                    assert host.noop_description is not None, "no noop description was set"
                    warnings.warn(
                        'No noop_description set for test: {0} (got "{1}")'.format(
                            op_test_name,
                            host.noop_description,
                        ),
                    )

    return TestTests


# Find available operation tests
operations = sorted(
    [
        filename
        for filename in listdir(path.join("tests", "operations"))
        if path.isdir(path.join("tests", "operations", filename))
    ],
)

# Generate the classes, attaching to locals so nose picks them up
for operation_name in operations:
    locals()[operation_name] = make_operation_tests(operation_name)



# Monkey patch the stdlib for gevent - this _must_ happen here, as early as
# possible, to avoid problems with the stdlib usage elsewhere in pyinfra. API
# scripts should also patch gevent at the start of their execution.
from gevent import monkey  # noqa

monkey.patch_all()  # noqa



from collections import defaultdict
from os import listdir, path
from types import GeneratorType
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union

from pyinfra import logger
from pyinfra.api.inventory import Inventory
from pyinfra.context import ctx_inventory

from .exceptions import CliError
from .util import exec_file, try_import_module_attribute

HostType = Union[str, Tuple[str, Dict]]

# Hosts in an inventory can be just the hostname or a tuple (hostname, data)
ALLOWED_HOST_TYPES = (str, tuple)


def _is_inventory_group(key: str, value: Any):
    """
    Verify that a module-level variable (key = value) is a valid inventory group.
    """

    if key.startswith("_") or not isinstance(value, (list, tuple, GeneratorType)):
        return False

    # If the group is a tuple of (hosts, data), check the hosts
    if isinstance(value, tuple):
        value = value[0]

    # Expand any generators of hosts
    if isinstance(value, GeneratorType):
        value = list(value)

    return all(isinstance(item, ALLOWED_HOST_TYPES) for item in value)


def _get_group_data(dirname: str):
    group_data = {}
    group_data_directory = path.join(dirname, "group_data")

    logger.debug("Checking possible group_data directory: %s", dirname)

    if path.exists(group_data_directory):
        files = listdir(group_data_directory)

        for file in files:
            if not file.endswith(".py"):
                continue

            group_data_file = path.join(group_data_directory, file)
            group_name = path.basename(file)[:-3]

            logger.debug("Looking for group data in: %s", group_data_file)

            # Read the files locals into a dict
            attrs = exec_file(group_data_file, return_locals=True)
            keys = attrs.get("__all__", attrs.keys())

            group_data[group_name] = {
                key: value
                for key, value in attrs.items()
                if key in keys and not key.startswith("_")
            }

    return group_data


def _get_groups_from_filename(inventory_filename: str):
    attrs = exec_file(inventory_filename, return_locals=True)

    return {key: value for key, value in attrs.items() if _is_inventory_group(key, value)}


T = TypeVar("T")


def _get_any_tuple_first(item: Union[T, Tuple[T, Any]]) -> T:
    return item[0] if isinstance(item, tuple) else item


def make_inventory(
    inventory: str,
    override_data=None,
    cwd: Optional[str] = None,
    group_data_directories=None,
):
    # First, try loading the inventory as if it's a Python import function
    try:
        inventory_func = try_import_module_attribute(inventory)
    except (CliError, ValueError):
        # If not an import, load as if from the filesystem *or* comma separated list, which also
        # loads any all.py group data files (imported functions do not load group data).
        return make_inventory_from_files(inventory, override_data, cwd, group_data_directories)
    else:
        return make_inventory_from_func(inventory_func, override_data)


def make_inventory_from_func(
    inventory_func: Callable[[], Dict[str, List[HostType]]],
    override_data: Optional[Dict[Any, Any]] = None,
):
    logger.warning("Loading inventory via import function is in alpha!")

    try:
        groups = inventory_func()
    except Exception as e:
        raise CliError(f"Failed to load inventory function: {inventory_func.__name__}: {e}")

    if not isinstance(groups, dict):
        raise TypeError(f"Inventory function {inventory_func.__name__} did not return a dictionary")

    # TODO: this shouldn't be required to make an inventory, groups should suffice
    combined_host_list = set()
    groups_with_data: Dict[str, Tuple[List[HostType], Dict]] = {}

    for key, hosts in groups.items():
        data: Dict = {}

        if isinstance(hosts, tuple):
            hosts, data = hosts

        if not isinstance(data, dict):
            raise TypeError(
                f"Inventory function {inventory_func.__name__} "
                f"group contains non-dictionary data: {key}"
            )

        for host in hosts:
            if not isinstance(host, ALLOWED_HOST_TYPES):
                raise TypeError(
                    f"Inventory function {inventory_func.__name__} invalid host: {host}"
                )

            host = _get_any_tuple_first(host)

            if not isinstance(host, str):
                raise TypeError(
                    f"Inventory function {inventory_func.__name__} invalid host name: {host}"
                )

            combined_host_list.add(host)

        groups_with_data[key] = (hosts, data)

    return Inventory(
        (list(combined_host_list), {}),
        override_data=override_data,
        **groups_with_data,
    )


def make_inventory_from_files(
    inventory_filename: str,
    override_data=None,
    cwd: Optional[str] = None,
    group_data_directories=None,
):
    """
    Builds a ``pyinfra.api.Inventory`` from the filesystem. If the file does not exist
    and doesn't contain a / attempts to use that as the only hostname.
    """

    file_groupname = None

    # TODO: this type is complex & convoluted, fix this
    groups: Dict[str, Union[List[str], Tuple[List[str], Dict[str, Any]]]]

    # If we're not a valid file we assume a list of comma separated hostnames
    if not path.exists(inventory_filename):
        groups = {
            "all": inventory_filename.split(","),
        }
    else:
        groups = _get_groups_from_filename(inventory_filename)
        # Used to set all the hosts to an additional group - that of the filename
        # ie inventories/dev.py means all the hosts are in the dev group, if not present
        file_groupname = path.basename(inventory_filename).rsplit(".", 1)[0]

    all_data: Dict[str, Any] = {}

    if "all" in groups:
        all_hosts = groups.pop("all")

        if isinstance(all_hosts, tuple):
            all_hosts, all_data = all_hosts

    # Build all out of the existing hosts if not defined
    else:
        all_hosts = []
        for hosts in groups.values():
            # Groups can be a list of hosts or tuple of (hosts, data)
            hosts = _get_any_tuple_first(hosts)

            for host in hosts:
                # Hosts can be a hostname or tuple of (hostname, data)
                hostname = _get_any_tuple_first(host)

                if hostname not in all_hosts:
                    all_hosts.append(hostname)

    groups["all"] = (all_hosts, all_data)

    # Apply the filename group if not already defined
    if file_groupname and file_groupname not in groups:
        groups[file_groupname] = all_hosts

    # In pyinfra an inventory is a combination of (hostnames + data). However, in CLI
    # mode we want to be define this in separate files (inventory / group data). The
    # issue is we want inventory access within the group data files - but at this point
    # we're not ready to make an Inventory. So here we just create a fake one, and
    # attach it to the inventory context while we import the data files.
    logger.debug("Creating fake inventory...")

    fake_groups = {
        # In API mode groups *must* be tuples of (hostnames, data)
        name: group if isinstance(group, tuple) else (group, {})
        for name, group in groups.items()
    }
    fake_inventory = Inventory((all_hosts, all_data), **fake_groups)

    possible_group_data_folders = []
    if cwd:
        possible_group_data_folders.append(cwd)
    inventory_dirname = path.abspath(path.dirname(inventory_filename))
    if inventory_dirname != cwd:
        possible_group_data_folders.append(inventory_dirname)

    if group_data_directories:
        possible_group_data_folders.extend(group_data_directories)

    group_data: Dict[str, Dict[str, Any]] = defaultdict(dict)

    with ctx_inventory.use(fake_inventory):
        for folder in possible_group_data_folders:
            for group_name, data in _get_group_data(folder).items():
                group_data[group_name].update(data)

    # For each group load up any data
    for name, hosts in groups.items():
        data = {}

        if isinstance(hosts, tuple):
            hosts, data = hosts

        if name in group_data:
            data.update(group_data.pop(name))

        # Attach to group object
        groups[name] = (hosts, data)

    # Loop back through any leftover group data and create an empty (for now)
    # group - this is because inventory @connectors can attach arbitrary groups
    # to hosts, so we need to support that.
    for name, data in group_data.items():
        groups[name] = ([], data)

    return Inventory(groups.pop("all"), override_data=override_data, **groups)



import shlex
from collections import defaultdict
from io import StringIO
from urllib.parse import urlparse

from pyinfra.facts.files import File
from pyinfra.facts.rpm import RpmPackage


def _package_name(package):
    if isinstance(package, list):
        return package[0]
    return package


def _has_package(package, packages, expand_package_fact=None, match_any=False):
    def in_packages(pkg_name, pkg_versions):
        if not pkg_versions:
            return pkg_name in packages
        return pkg_name in packages and any(
            version in packages[pkg_name] for version in pkg_versions
        )

    packages_to_check = [package]
    if expand_package_fact:
        packages_to_check = expand_package_fact(package) or packages_to_check

    package_name_to_versions = defaultdict(set)
    for pkg in packages_to_check:
        if isinstance(pkg, list):
            package_name_to_versions[pkg[0]].add(pkg[1])
        else:
            package_name_to_versions[pkg]  # just make sure it exists

    checks = (
        in_packages(pkg_name, pkg_versions)
        for pkg_name, pkg_versions in package_name_to_versions.items()
    )

    if match_any:
        return any(checks), package_name_to_versions
    return all(checks), package_name_to_versions


def ensure_packages(
    host,
    packages,
    current_packages,
    present,
    install_command,
    uninstall_command,
    latest=False,
    upgrade_command=None,
    version_join=None,
    expand_package_fact=None,
):
    """
    Handles this common scenario:

    + We have a list of packages(/versions) to ensure
    + We have a map of existing package -> versions
    + We have the common command bits (install, uninstall, version "joiner")
    + Outputs commands to ensure our desired packages/versions
    + Optionally upgrades packages w/o specified version when present

    Args:
        packages (list): list of packages or package/versions
        current_packages (fact): fact returning dict of package names -> version
        present (bool): whether packages should exist or not
        install_command (str): command to prefix to list of packages to install
        uninstall_command (str): as above for uninstalling packages
        latest (bool): whether to upgrade installed packages when present
        upgrade_command (str): as above for upgrading
        version_join (str): the package manager specific "joiner", ie ``=`` for \
            ``<apt_pkg>=<version>``
    """

    if packages is None:
        return

    if isinstance(packages, str):
        packages = [packages]

    if version_join:
        packages = [
            package[0] if len(package) == 1 else package
            for package in [package.rsplit(version_join, 1) for package in packages]
        ]

    diff_packages = []
    diff_expanded_packages = {}

    upgrade_packages = []

    if present is True:
        for package in packages:
            has_package, expanded_packages = _has_package(
                package,
                current_packages,
                expand_package_fact,
            )

            if not has_package:
                diff_packages.append(package)
                diff_expanded_packages[_package_name(package)] = expanded_packages
            else:
                # Present packages w/o version specified - for upgrade if latest
                if isinstance(package, str):
                    upgrade_packages.append(package)

                if not latest:
                    pkg_name = _package_name(package)
                    if pkg_name in current_packages:
                        host.noop(
                            "package {0} is installed ({1})".format(
                                package,
                                ", ".join(current_packages[pkg_name]),
                            ),
                        )
                    else:
                        host.noop("package {0} is installed".format(package))

    if present is False:
        for package in packages:
            # String version, just check if existing
            has_package, expanded_packages = _has_package(
                package,
                current_packages,
                expand_package_fact,
                match_any=True,
            )

            if has_package:
                diff_packages.append(package)
                diff_expanded_packages[_package_name(package)] = expanded_packages
            else:
                host.noop("package {0} is not installed".format(package))

    if diff_packages:
        command = install_command if present else uninstall_command

        joined_packages = [
            version_join.join(package) if isinstance(package, list) else package
            for package in diff_packages
        ]

        yield "{0} {1}".format(
            command,
            " ".join([shlex.quote(pkg) for pkg in joined_packages]),
        )

        for package in diff_packages:  # add/remove from current packages
            pkg_name = _package_name(package)
            version = "unknown"
            if isinstance(package, list):
                version = package[1]

            if present:
                current_packages[pkg_name] = [version]
                current_packages.update(diff_expanded_packages.get(pkg_name, {}))
            else:
                current_packages.pop(pkg_name, None)
                for name in diff_expanded_packages.get(pkg_name, {}):
                    current_packages.pop(name, None)

    if latest and upgrade_command and upgrade_packages:
        yield "{0} {1}".format(
            upgrade_command,
            " ".join([shlex.quote(pkg) for pkg in upgrade_packages]),
        )


def ensure_rpm(state, host, files, source, present, package_manager_command):
    original_source = source

    # If source is a url
    if urlparse(source).scheme:
        # Generate a temp filename (with .rpm extension to please yum)
        temp_filename = "{0}.rpm".format(state.get_temp_filename(source))

        # Ensure it's downloaded
        yield from files.download(source, temp_filename)

        # Override the source with the downloaded file
        source = temp_filename

    # Check for file .rpm information
    info = host.get_fact(RpmPackage, name=source)
    exists = False

    # We have info!
    if info:
        current_package = host.get_fact(RpmPackage, name=info["name"])
        if current_package and current_package["version"] == info["version"]:
            exists = True

    # Package does not exist and we want?
    if present and not exists:
        # If we had info, always install
        if info:
            yield "rpm -i {0}".format(source)
            host.create_fact(RpmPackage, kwargs={"name": info["name"]}, data=info)

        # This happens if we download the package mid-deploy, so we have no info
        # but also don't know if it's installed. So check at runtime, otherwise
        # the install will fail.
        else:
            yield "rpm -q `rpm -qp {0}` 2> /dev/null || rpm -i {0}".format(source)

    # Package exists but we don't want?
    elif exists and not present:
        yield "{0} remove -y {1}".format(package_manager_command, info["name"])
        host.delete_fact(RpmPackage, kwargs={"name": info["name"]})

    else:
        host.noop(
            "rpm {0} is {1}".format(
                original_source,
                "installed" if present else "not installed",
            ),
        )


def ensure_yum_repo(
    state,
    host,
    files,
    name_or_url,
    baseurl,
    present,
    description,
    enabled,
    gpgcheck,
    gpgkey,
    repo_directory="/etc/yum.repos.d/",
    type_=None,
):
    url = None
    url_parts = urlparse(name_or_url)
    if url_parts.scheme:
        url = name_or_url
        name_or_url = url_parts.path.split("/")[-1]
        if name_or_url.endswith(".repo"):
            name_or_url = name_or_url[:-5]

    filename = "{0}{1}.repo".format(repo_directory, name_or_url)

    # If we don't want the repo, just remove any existing file
    if not present:
        yield from files.file(filename, present=False)
        return

    # If we're a URL, download the repo if it doesn't exist
    if url:
        if not host.get_fact(File, path=filename):
            yield from files.download(url, filename)
        return

    # Description defaults to name
    description = description or name_or_url

    # Build the repo file from string
    repo_lines = [
        "[{0}]".format(name_or_url),
        "name={0}".format(description),
        "baseurl={0}".format(baseurl),
        "enabled={0}".format(1 if enabled else 0),
        "gpgcheck={0}".format(1 if gpgcheck else 0),
    ]

    if type_:
        repo_lines.append("type={0}".format(type_))

    if gpgkey:
        repo_lines.append("gpgkey={0}".format(gpgkey))

    repo_lines.append("")
    repo = "\n".join(repo_lines)
    repo = StringIO(repo)

    # Ensure this is the file on the server
    yield from files.put(repo, filename)



from pyinfra.api import FactBase

from .util.packaging import parse_packages

PACMAN_REGEX = r"^([0-9a-zA-Z\-]+)\s([0-9\._+a-z\-]+)"


class PacmanUnpackGroup(FactBase):
    """
    Returns a list of actual packages belonging to the provided package name,
    expanding groups or virtual packages.

    .. code:: python

        [
            "package_name",
        ]
    """

    requires_command = "pacman"

    default = list

    def command(self, name):
        # Accept failure here (|| true) for invalid/unknown packages
        return 'pacman -S --print-format "%n" {0} || true'.format(name)

    def process(self, output):
        return output


class PacmanPackages(FactBase):
    """
    Returns a dict of installed pacman packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "pacman -Q"
    requires_command = "pacman"

    default = dict

    def process(self, output):
        return parse_packages(PACMAN_REGEX, output)



import re

from pyinfra.api import FactBase


class SnapBaseFact(FactBase):
    abstract = True
    requires_command = "snap"


class SnapPackage(SnapBaseFact):
    """
    Returns information for an installed snap package

    .. code:: python

        {
            "name": "lxd",
            "publisher": "Canonical✓",
            "snap-id": "J60k4JY0HppjwOjW8dZdYc8obXKxujRu",
            "channel": "4.0/stable",
            "version": "4.0.9"
        }
    """

    default = dict
    _regexes = {
        "name": "^name:[ ]+(.*)",
        "publisher": r"^publisher:[ ]+(.*)",
        "snap-id": r"^snap-id:[ ]+(.*)",
        "channel": r"^tracking:[ ]+([\w\d.-]+/[\w\d.-]+)[/]?.*$",
        "version": r"^installed:[ ]+([\w\d.-]+).*$",
    }

    def command(self, package):
        return f"snap info {package}"

    def process(self, output):
        data = {}
        for line in output:
            for regex_name, regex in self._regexes.items():
                matches = re.match(regex, line)
                if matches:
                    data[regex_name] = matches.group(1)

        return data


class SnapPackages(SnapBaseFact):
    """
    Returns a list of installed snap packages:

    .. code:: python

        [
            "core",
            "core18",
            "core20",
            "lxd",
            "snapd"
        ]
    """

    default = list
    command = "snap list"

    def process(self, output):
        # Discard header output line from snap list command
        # 'snap list' command example output lines:
        #    $ snap list
        #    Name              Version        Rev    Tracking         Publisher     Notes
        #    core              16-2.57.2      13886  latest/stable    canonical✓    core
        #    lxd               4.0.9          22753  4.0/stable/…     canonical✓    -
        #
        return [snap.split()[0] for snap in output[1:]]



"""
Manage Ruby gem packages. (see https://rubygems.org/ )
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.gem import GemPackages

from .util.packaging import ensure_packages


@operation
def packages(packages=None, present=True, latest=False):
    """
    Add/remove/update gem packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.

    **Example:**

    .. code:: python

        # Note: Assumes that 'gem' is installed.
        gem.packages(
            name="Install rspec",
            packages=["rspec"],
        )
    """

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(GemPackages),
        present,
        install_command="gem install",
        uninstall_command="gem uninstall",
        upgrade_command="gem update",
        version_join=":",
        latest=latest,
    )



from pyinfra.api import operation


@operation(is_idempotent=False)
def agent(server=None, port=None):
    """
    Run puppet agent

    + server: master server URL
    + port: puppet master port

    Note: Either 'USE_SUDO_LOGIN=True' or 'USE_SU_LOGIN=True'
    for puppet.agent() as `puppet` is added to the path in
    the .bash_profile.

    **Example:**

    .. code:: python

        puppet.agent()

        # We also expect a return code of:
        # 0=no changes or 2=changes applied
        puppet.agent(
            name="Run the puppet agent",
            success_exit_codes=[0, 2],
        )

    """

    args = []

    if server:
        args.append("--server=%s" % server)
    if port:
        args.append("--masterport=%s" % port)

    yield "puppet agent -t %s" % " ".join(args)



import json
import warnings
from importlib import import_module
from os import listdir, path
from unittest import TestCase

from pyinfra.api import StringCommand
from pyinfra.api.facts import ShortFactBase
from pyinfra_cli.util import json_encode

from .util import JsonTest, get_command_string

# show full diff on json
TestCase.maxDiff = None


def _make_command(command_attribute, args):
    if callable(command_attribute):
        if not isinstance(args, list):
            args = [args]

        return command_attribute(*args)
    return command_attribute


def make_fact_tests(folder_name):
    module_name, fact_name = folder_name.split(".")
    module = import_module("pyinfra.facts.{0}".format(module_name))
    fact = getattr(module, fact_name)()

    class TestTests(TestCase, metaclass=JsonTest):
        jsontest_files = path.join("tests", "facts", folder_name)
        jsontest_prefix = "test_{0}_".format(fact.name)

        def jsontest_function(self, test_name, test_data, fact=fact):
            short_fact = None

            if isinstance(fact, ShortFactBase):
                short_fact = fact
                fact = fact.fact()

            test_args = test_data.get("arg", [])
            command = _make_command(fact.command, test_args)

            if "command" in test_data:
                assert get_command_string(StringCommand(command)) == test_data["command"]
            else:
                warnings.warn(
                    'No command set for test: {0} (got "{1}")'.format(
                        test_name,
                        command,
                    ),
                )

            requires_command = _make_command(fact.requires_command, test_args)

            if requires_command:
                if "requires_command" in test_data:
                    assert requires_command == test_data["requires_command"]
                else:
                    warnings.warn(
                        'No requires command set for test: {0} (got "{1}")'.format(
                            test_name,
                            requires_command,
                        ),
                    )

            data = fact.process(test_data["output"])
            if short_fact:
                data = short_fact.process_data(data)

            # Encode/decode data to ensure datetimes/etc become JSON
            data = json.loads(json.dumps(data, default=json_encode))
            try:
                assert data == test_data["fact"]
            except AssertionError as e:
                print()
                print("--> GOT:\n", json.dumps(data, indent=4, default=json_encode))
                print(
                    "--> WANT:",
                    json.dumps(
                        test_data["fact"],
                        indent=4,
                        default=json_encode,
                    ),
                )
                raise e

    TestTests.__name__ = "Fact{0}".format(fact.name)
    return TestTests


# Find available fact tests
fact_tests = sorted(
    [
        filename
        for filename in listdir(path.join("tests", "facts"))
        if path.isdir(path.join("tests", "facts", filename))
    ],
)

# Generate the classes, attaching to local
for fact_name in fact_tests:
    locals()[fact_name] = make_fact_tests(fact_name)



"""
Deploys come in two forms: on-disk, eg deploy.py, and @deploy wrapped functions.
The latter enable re-usable (across CLI and API based execution) pyinfra extension
creation (eg pyinfra-openstack).
"""

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Union

import pyinfra
from pyinfra import context, logger
from pyinfra.context import ctx_host, ctx_state

from .arguments import pop_global_arguments
from .exceptions import PyinfraError
from .host import Host
from .util import get_args_kwargs_spec, get_call_location, memoize

if TYPE_CHECKING:
    from pyinfra.api.state import State


@memoize
def show_state_host_arguments_warning(call_location):
    logger.warning(
        (
            "{0}:\n\tLegacy deploy function detected! Deploys should no longer define "
            "`state` and `host` arguments."
        ).format(call_location),
    )


def add_deploy(state: "State", deploy_func: Callable[..., Any], *args, **kwargs):
    """
    Prepare & add an deploy to pyinfra.state by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        deploy_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    """

    if pyinfra.is_cli:
        raise PyinfraError(
            (
                "`add_deploy` should not be called when pyinfra is executing in CLI mode! ({0})"
            ).format(get_call_location()),
        )

    hosts = kwargs.pop("host", state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    with ctx_state.use(state):
        for deploy_host in hosts:
            with ctx_host.use(deploy_host):
                deploy_func(*args, **kwargs)


def deploy(func_or_name: Union[Callable[..., Any], str], data_defaults=None, _call_location=None):
    """
    Decorator that takes a deploy function (normally from a pyinfra_* package)
    and wraps any operations called inside with any deploy-wide kwargs/data.
    """

    # If not decorating, return function with config attached
    if isinstance(func_or_name, str):
        name = func_or_name

        def decorator(f):
            f.deploy_name = name
            if data_defaults:
                f.deploy_data = data_defaults
            return deploy(f, _call_location=get_call_location())

        return decorator

    # Actually decorate!
    func = func_or_name

    # Check whether an operation is "legacy" - ie contains state=None, host=None kwargs
    # TODO: remove this in v3
    is_legacy = False
    args, kwargs = get_args_kwargs_spec(func)
    if all(key in kwargs and kwargs[key] is None for key in ("state", "host")):
        show_state_host_arguments_warning(_call_location or get_call_location())
        is_legacy = True
    func.is_legacy = is_legacy  # type: ignore

    @wraps(func)
    def decorated_func(*args, **kwargs):
        deploy_kwargs, _ = pop_global_arguments(kwargs)

        # If this is a legacy operation function (ie - state & host arg kwargs), ensure that state
        # and host are included as kwargs.
        if func.is_legacy:
            if "state" not in kwargs:
                kwargs["state"] = context.state
            if "host" not in kwargs:
                kwargs["host"] = context.host
        # If not legacy, pop off any state/host kwargs that may come from legacy @deploy functions
        else:
            kwargs.pop("state", None)
            kwargs.pop("host", None)

        # Name the deploy
        deploy_name = getattr(func, "deploy_name", func.__name__)
        deploy_data = getattr(func, "deploy_data", None)

        with context.host.deploy(
            name=deploy_name,
            kwargs=deploy_kwargs,
            data=deploy_data,
        ):
            return func(*args, **kwargs)

    return decorated_func



"""
.. warning::
    This connector is in alpha and may change in future releases.

Generate one or more SSH hosts from a Terraform output variable. The variable
must be a list of hostnames or IP addresses that ``pyinfra`` can connect to
over SSH. Currently there is no support for specifying SSH user/pass/port/key
from Terraform, these must be provided via ``pyinfra`` group data or ``--data``
CLI flags.

Output is fetched from a flattened JSON dictionary output from ``terraform output
-json``. For example the following object:

.. code:: json

    {
      "server_group": {
        "value": {
          "server_group_node_ips": [
            "1.2.3.4",
            "1.2.3.5",
            "1.2.3.6"
          ]
        }
      }
    }

The IP list ``server_group_node_ips`` would be used like so:

.. code:: python

    pyinfra @terraform/server_group.value.server_group_node_ips ...
"""

import json

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize
from pyinfra.progress import progress_spinner


@memoize
def show_warning():
    logger.warning("The @terraform connector is in alpha!")


def _flatten_dict_gen(d, parent_key, sep):
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            yield from _flatten_dict(v, new_key, sep=sep).items()
        else:
            yield new_key, v


def _flatten_dict(d: dict, parent_key: str = "", sep: str = "."):
    return dict(_flatten_dict_gen(d, parent_key, sep))


def make_names_data(output_key=None):
    show_warning()

    if not output_key:
        raise InventoryError("No Terraform output key!")

    with progress_spinner({"fetch terraform output"}):
        tf_output_raw = local.shell("terraform output -json")

    tf_output = json.loads(tf_output_raw)
    tf_output = _flatten_dict(tf_output)

    tf_output_value = tf_output.get(output_key)
    if tf_output_value is None:
        raise InventoryError(f"No Terraform output with key: `{output_key}`")

    if not isinstance(tf_output_value, list):
        raise InventoryError(
            "Invalid Terraform output type, should be `list`, got "
            f"`{type(tf_output_value).__name__}`",
        )

    for ssh_target in tf_output_value:
        if isinstance(ssh_target, dict):
            name = ssh_target.pop("name", ssh_target.get("ssh_hostname"))
            if name is None:
                raise InventoryError(
                    "Invalid Terraform list item, missing `name` or `ssh_hostname` keys",
                )
            yield f"@terraform/{name}", ssh_target, ["@terraform"]

        elif isinstance(ssh_target, str):
            data = {"ssh_hostname": ssh_target}
            yield f"@terraform/{ssh_target}", data, ["@terraform"]

        else:
            raise InventoryError(
                "Invalid Terraform list item, should be `dict` or `str` got "
                f"`{type(ssh_target).__name__}`",
            )



from unittest import TestCase

try:
    from pathlib import PurePosixPath

    HAS_PATHLIB = True
except ImportError:
    HAS_PATHLIB = False

import pytest

from pyinfra.api import (
    FileDownloadCommand,
    FileUploadCommand,
    FunctionCommand,
    MaskString,
    QuoteString,
    RsyncCommand,
    StringCommand,
)
from pyinfra.api.command import PyinfraCommand


class TestBaseCommand(TestCase):
    def test_base_command_no_execute(self):
        cmd = PyinfraCommand()
        with self.assertRaises(NotImplementedError):
            cmd.execute(None, None, None)


class TestStringCommand(TestCase):
    def test_normal(self):
        cmd = StringCommand("hello", "world")
        assert str(cmd) == cmd.get_raw_value() == "hello world"

    def test_masked(self):
        cmd = StringCommand(MaskString("adsfg"))
        assert cmd.get_raw_value() == "adsfg"
        assert str(cmd) == "***"

    def test_mixed_masked(self):
        cmd = StringCommand("some", "stuff", MaskString("mask me"), "other", "stuff")
        assert cmd.get_raw_value() == "some stuff mask me other stuff"
        assert str(cmd) == "some stuff *** other stuff"

    def test_nested(self):
        nested_cmd = StringCommand("some", "stuff")
        cmd = StringCommand("hello", nested_cmd, "world")
        assert str(cmd) == cmd.get_raw_value() == "hello some stuff world"

    def test_quote(self):
        quoted_str = QuoteString("quote me!")
        cmd = StringCommand("hello", quoted_str)
        assert str(cmd) == cmd.get_raw_value() == "hello 'quote me!'"

    def test_eq(self):
        assert StringCommand("hello world") == StringCommand("hello world")
        with self.assertRaises(AssertionError):
            assert StringCommand("a") == StringCommand("b")

    def test_separator(self):
        cmd = StringCommand("hello world", "yes", _separator="")
        assert str(cmd) == "hello worldyes"

    def test_int(self):
        cmd = StringCommand("hello", "world", 4420)
        assert str(cmd) == "hello world 4420"

    def test_list(self):
        cmd = StringCommand("hello", "world", ["a", "b", "c"])
        assert str(cmd) == "hello world ['a', 'b', 'c']"

    @pytest.mark.skipif(
        not HAS_PATHLIB,
        reason="Requires Python 3.4+ (pathlib module)",
    )
    def test_path(self):
        cmd = StringCommand("hello", "world", PurePosixPath("/path/to/somewhere"))
        assert str(cmd) == "hello world /path/to/somewhere"


class TestFileCommands(TestCase):
    def test_file_upload_command_repr(self):
        cmd = FileUploadCommand("src", "dest")
        assert repr(cmd) == "FileUploadCommand(src, dest)"

    def test_file_download_command_repr(self):
        cmd = FileDownloadCommand("src", "dest")
        assert repr(cmd) == "FileDownloadCommand(src, dest)"

    def test_rsync_command_repr(self):
        cmd = RsyncCommand("src", "dest", ["-a"])
        assert repr(cmd) == "RsyncCommand(src, dest, ['-a'])"


class TestFunctionCommand(TestCase):
    def test_function_command_repr(self):
        def some_function():
            pass

        cmd = FunctionCommand(some_function, (), {})
        assert repr(cmd) == "FunctionCommand(some_function, (), {})"



import math
import os
import platform
import sys
from collections import deque
from contextlib import contextmanager
from threading import Event, Thread
from time import sleep

import pyinfra

IS_WINDOWS = platform.system() == "Windows"

WAIT_TIME = 1 / 5
WAIT_CHARS = deque(("-", "/", "|", "\\"))

# Hacky way of getting terminal size (so can clear lines)
# Source: http://stackoverflow.com/questions/566746
IS_TTY = sys.stdout.isatty() and sys.stderr.isatty()
TERMINAL_WIDTH = 0

if IS_TTY:
    try:
        TERMINAL_WIDTH = os.get_terminal_size().columns
    except AttributeError:
        if not IS_WINDOWS:
            terminal_size = os.popen("stty size", "r").read().split()
            if len(terminal_size) == 2:
                TERMINAL_WIDTH = int(terminal_size[1])


def _print_spinner(stop_event, progress_queue):
    if not IS_TTY or os.environ.get("PYINFRA_PROGRESS") == "off":
        return

    progress = ""
    text = ""

    while True:
        # Stop when asked too
        if stop_event.is_set():
            break

        WAIT_CHARS.rotate(1)

        try:
            progress = progress_queue[-1]
        except IndexError:
            pass

        text = "    {0}".format(
            " ".join((WAIT_CHARS[0], progress)),
        )
        text = "{0}\r".format(text)

        sys.stderr.write(text)
        sys.stderr.flush()

        # In pyinfra_cli's __main__ we set stdout & stderr to be line buffered,
        # so write this escape code (clear line) into the buffer but don't flush,
        # such that any next print/log/etc clear the line first.
        if not IS_WINDOWS:
            sys.stderr.write("\033[K")

        sleep(WAIT_TIME)


@contextmanager
def progress_spinner(items, prefix_message=None):
    # If there's no current state context we're not in CLI mode, so just return a noop
    # handler and exit.
    if not pyinfra.is_cli:
        yield lambda complete_item: None
        return

    if not isinstance(items, set):
        items = set(items)

    total_items = len(items)
    stop_event = Event()

    def make_progress_message(include_items=True):
        message_bits = []

        # If we only have 1 item, don't show %
        if total_items > 1:
            percentage_complete = 0
            complete = total_items - len(items)
            percentage_complete = int(math.floor(complete / total_items * 100))
            message_bits.append(
                "{0}% ({1}/{2})".format(
                    percentage_complete,
                    complete,
                    total_items,
                ),
            )

        if prefix_message:
            message_bits.append(prefix_message)

        if include_items and items:
            # Plus 3 for the " - " joining below
            message_length = sum((len(message) + 3) for message in message_bits)
            # -8 for padding left+right, -2 for {} wrapping
            items_allowed_width = TERMINAL_WIDTH - 10 - message_length

            if items_allowed_width > 0:
                items_string = "{%s}" % (", ".join("{0}".format(i) for i in items))
                if len(items_string) >= items_allowed_width:
                    items_string = "%s...}" % (
                        # -3 for the ...
                        items_string[: items_allowed_width - 3],
                    )

                message_bits.append(items_string)

        return " - ".join(message_bits)

    progress_queue = deque((make_progress_message(),))

    def progress(complete_item):
        if complete_item not in items:
            raise ValueError(
                "Invalid complete item: {0} not in {1}".format(
                    complete_item,
                    items,
                ),
            )

        items.remove(complete_item)
        progress_queue.append(make_progress_message())

    # Kick off the spinner thread
    spinner_thread = Thread(
        target=_print_spinner,
        args=(stop_event, progress_queue),
    )
    spinner_thread.daemon = True
    spinner_thread.start()

    # Yield allowing the actual code the spinner waits for to run
    yield progress

    # Finally, stop the spinner
    stop_event.set()
    spinner_thread.join()



import json
import os
from datetime import datetime
from importlib import import_module
from io import IOBase
from os import path
from pathlib import Path
from types import FunctionType, ModuleType
from typing import TYPE_CHECKING, Callable

import click
import gevent

from pyinfra import logger, state
from pyinfra.api.command import PyinfraCommand
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.host import HostData
from pyinfra.api.operation import OperationMeta
from pyinfra.context import ctx_config, ctx_host
from pyinfra.progress import progress_spinner

from .exceptions import CliError, UnexpectedExternalError

if TYPE_CHECKING:
    from pyinfra.api.state import State

# Cache for compiled Python deploy code
PYTHON_CODES = {}


def is_subdir(child, parent):
    child = path.realpath(child)
    parent = path.realpath(parent)
    relative = path.relpath(child, start=parent)
    return not relative.startswith(os.pardir)


def exec_file(filename, return_locals: bool = False, is_deploy_code: bool = False):
    """
    Execute a Python file and optionally return it's attributes as a dict.
    """

    old_current_exec_filename = state.current_exec_filename
    state.current_exec_filename = filename

    if filename not in PYTHON_CODES:
        with open(filename, "r", encoding="utf-8") as f:
            code = f.read()

        code = compile(code, filename, "exec")
        PYTHON_CODES[filename] = code

    # Create some base attributes for our "module"
    data = {
        "__file__": filename,
    }

    # Execute the code with locals/globals going into the dict above
    try:
        exec(PYTHON_CODES[filename], data)
    except Exception as e:
        if isinstance(e, (PyinfraError, UnexpectedExternalError)):
            raise
        raise UnexpectedExternalError(e, filename)

    state.current_exec_filename = old_current_exec_filename
    return data


def json_encode(obj):
    # pyinfra types
    if isinstance(obj, HostData):
        return obj.dict()

    if isinstance(obj, PyinfraCommand):
        return repr(obj)

    if isinstance(obj, OperationMeta):
        return repr(obj)

    # Python types
    if isinstance(obj, ModuleType):
        return "Module: {0}".format(obj.__name__)

    if isinstance(obj, FunctionType):
        return "Function: {0}".format(obj.__name__)

    if isinstance(obj, datetime):
        return obj.isoformat()

    if isinstance(obj, IOBase):
        if hasattr(obj, "name"):
            return "File: {0}".format(obj.name)

        if hasattr(obj, "template"):
            return "Template: {0}".format(obj.template)

        obj.seek(0)
        return "In memory file: {0}".format(obj.read())

    if isinstance(obj, Path):
        return str(obj)

    if isinstance(obj, set):
        return sorted(list(obj))

    if isinstance(obj, bytes):
        return obj.decode()

    raise TypeError("Cannot serialize: {0} ({1})".format(type(obj), obj))


def parse_cli_arg(arg):
    if isinstance(arg, list):
        return [parse_cli_arg(a) for a in arg]

    if arg.lower() == "false":
        return False

    if arg.lower() == "true":
        return True

    try:
        return int(arg)
    except (TypeError, ValueError):
        pass

    try:
        return json.loads(arg)
    except ValueError:
        pass

    return arg


def try_import_module_attribute(path, prefix=None):
    mod_path, attr_name = path.rsplit(".", 1)
    module = None

    if prefix:
        full_path = f"{prefix}.{mod_path}"
        try:
            module = import_module(full_path)
        except (ModuleNotFoundError, ImportError):
            pass
    else:
        full_path = mod_path

    if module is None:
        try:
            module = import_module(mod_path)
        except (ModuleNotFoundError, ImportError):
            raise CliError(f"No such module: {full_path}")

    attr = getattr(module, attr_name, None)
    if attr is None:
        raise CliError(f"No such attribute in module {full_path}: {attr_name}")

    return attr


def _parallel_load_hosts(state: "State", callback: Callable, name: str):
    def load_file(local_host):
        try:
            with ctx_config.use(state.config.copy()):
                with ctx_host.use(local_host):
                    callback()
                    logger.info(
                        "{0}{1} {2}".format(
                            local_host.print_prefix,
                            click.style("Ready:", "green"),
                            click.style(name, bold=True),
                        ),
                    )
        except Exception as e:
            return e

    greenlet_to_host = {
        state.pool.spawn(load_file, host): host for host in state.inventory.iter_active_hosts()
    }

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            result = greenlet.get()
            if isinstance(result, Exception):
                raise result
            progress(host)


def load_deploy_file(state: "State", filename):
    state.current_deploy_filename = filename
    _parallel_load_hosts(state, lambda: exec_file(filename), filename)


def load_func(state: "State", op_func, *args, **kwargs):
    _parallel_load_hosts(state, lambda: op_func(*args, **kwargs), op_func.__name__)



"""
This file as originally part of the "sshuserclient" pypi package. The GitHub
source has now vanished (https://github.com/tobald/sshuserclient).
"""

from os import path

from gevent.lock import BoundedSemaphore
from paramiko import (
    HostKeys,
    MissingHostKeyPolicy,
    ProxyCommand,
    SSHClient as ParamikoClient,
    SSHException,
)
from paramiko.agent import AgentRequestHandler

from pyinfra import logger
from pyinfra.api.util import memoize

from .config import SSHConfig

HOST_KEYS_LOCK = BoundedSemaphore()


class StrictPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        logger.error("No host key for {0} found in known_hosts".format(hostname))
        raise SSHException(
            "StrictPolicy: No host key for {0} found in known_hosts".format(hostname),
        )


class AcceptNewPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        logger.warning(
            (
                f"No host key for {hostname} found in known_hosts, "
                "accepting & adding to host keys file"
            ),
        )

        with HOST_KEYS_LOCK:
            host_keys = client.get_host_keys()
            host_keys.add(hostname, key.get_name(), key)
            # The paramiko client saves host keys incorrectly whereas the host keys object does
            # this correctly, so use that with the client filename variable.
            # See: https://github.com/paramiko/paramiko/pull/1989
            host_keys.save(client._host_keys_filename)


class AskPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        should_continue = input(
            "No host key for {0} found in known_hosts, do you want to continue [y/n] ".format(
                hostname,
            ),
        )
        if should_continue.lower() != "y":
            raise SSHException(
                "AskPolicy: No host key for {0} found in known_hosts".format(hostname),
            )
        with HOST_KEYS_LOCK:
            host_keys = client.get_host_keys()
            host_keys.add(hostname, key.get_name(), key)
            # The paramiko client saves host keys incorrectly whereas the host keys object does
            # this correctly, so use that with the client filename variable.
            # See: https://github.com/paramiko/paramiko/pull/1989
            host_keys.save(client._host_keys_filename)
        logger.warning("Added host key for {0} to known_hosts".format(hostname))
        return


class WarningPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        logger.warning("No host key for {0} found in known_hosts".format(hostname))


def get_missing_host_key_policy(policy):
    if policy is None or policy == "ask":
        return AskPolicy()
    if policy == "no" or policy == "off":
        return WarningPolicy()
    if policy == "yes":
        return StrictPolicy()
    if policy == "accept-new":
        return AcceptNewPolicy()
    raise SSHException("Invalid value StrictHostKeyChecking={}".format(policy))


@memoize
def get_ssh_config(user_config_file=None):
    logger.debug("Loading SSH config: %s", user_config_file)

    if user_config_file is None:
        user_config_file = path.expanduser("~/.ssh/config")

    if path.exists(user_config_file):
        with open(user_config_file, encoding="utf-8") as f:
            ssh_config = SSHConfig()
            ssh_config.parse(f)
            return ssh_config


@memoize
def get_host_keys(filename):
    with HOST_KEYS_LOCK:
        host_keys = HostKeys()

        try:
            host_keys.load(filename)
        # When paramiko encounters a bad host keys line it sometimes bails the
        # entire load incorrectly.
        # See: https://github.com/paramiko/paramiko/pull/1990
        except Exception as e:
            logger.warning("Failed to load host keys from {0}: {1}".format(filename, e))

        return host_keys


class SSHClient(ParamikoClient):
    """
    An SSHClient which honors ssh_config and supports proxyjumping
    original idea at http://bitprophet.org/blog/2012/11/05/gateway-solutions/.
    """

    def connect(
        self,
        hostname,
        _pyinfra_ssh_forward_agent=None,
        _pyinfra_ssh_config_file=None,
        _pyinfra_ssh_known_hosts_file=None,
        _pyinfra_ssh_strict_host_key_checking=None,
        _pyinfra_ssh_paramiko_connect_kwargs=None,
        **kwargs,
    ):
        (
            hostname,
            config,
            forward_agent,
            missing_host_key_policy,
            host_keys_file,
        ) = self.parse_config(
            hostname,
            kwargs,
            ssh_config_file=_pyinfra_ssh_config_file,
            strict_host_key_checking=_pyinfra_ssh_strict_host_key_checking,
        )
        self.set_missing_host_key_policy(missing_host_key_policy)
        config.update(kwargs)

        if _pyinfra_ssh_known_hosts_file:
            host_keys_file = _pyinfra_ssh_known_hosts_file

        # Overwrite paramiko empty defaults with @memoize-d host keys object
        self._host_keys = get_host_keys(host_keys_file)
        self._host_keys_filename = host_keys_file

        if _pyinfra_ssh_paramiko_connect_kwargs:
            config.update(_pyinfra_ssh_paramiko_connect_kwargs)

        self._ssh_config = config
        super().connect(hostname, **config)

        if _pyinfra_ssh_forward_agent is not None:
            forward_agent = _pyinfra_ssh_forward_agent

        if forward_agent:
            # Enable SSH forwarding
            session = self.get_transport().open_session()
            AgentRequestHandler(session)

    def gateway(self, hostname, host_port, target, target_port):
        transport = self.get_transport()
        return transport.open_channel(
            "direct-tcpip",
            (target, target_port),
            (hostname, host_port),
        )

    def parse_config(
        self,
        hostname,
        initial_cfg=None,
        ssh_config_file=None,
        strict_host_key_checking=None,
    ):
        cfg = {"port": 22}
        cfg.update(initial_cfg or {})

        forward_agent = False
        missing_host_key_policy = get_missing_host_key_policy(strict_host_key_checking)
        host_keys_file = path.expanduser("~/.ssh/known_hosts")  # OpenSSH default

        ssh_config = get_ssh_config(ssh_config_file)
        if not ssh_config:
            return hostname, cfg, forward_agent, missing_host_key_policy, host_keys_file

        host_config = ssh_config.lookup(hostname)
        forward_agent = host_config.get("forwardagent") == "yes"

        # If not overridden, apply any StrictHostKeyChecking
        if strict_host_key_checking is None and "stricthostkeychecking" in host_config:
            missing_host_key_policy = get_missing_host_key_policy(
                host_config["stricthostkeychecking"],
            )

        if "userknownhostsfile" in host_config:
            host_keys_file = path.expanduser(host_config["userknownhostsfile"])

        if "hostname" in host_config:
            hostname = host_config["hostname"]

        if "user" in host_config:
            cfg["username"] = host_config["user"]

        if "identityfile" in host_config:
            cfg["key_filename"] = host_config["identityfile"]

        if "port" in host_config:
            cfg["port"] = int(host_config["port"])

        if "proxycommand" in host_config:
            cfg["sock"] = ProxyCommand(host_config["proxycommand"])

        elif "proxyjump" in host_config:
            hops = host_config["proxyjump"].split(",")
            sock = None

            for i, hop in enumerate(hops):
                hop_hostname, hop_config = self.derive_shorthand(ssh_config, hop)
                logger.debug("SSH ProxyJump through %s:%s", hop_hostname, hop_config["port"])

                c = SSHClient()
                c.connect(
                    hop_hostname, _pyinfra_ssh_config_file=ssh_config_file, sock=sock, **hop_config
                )

                if i == len(hops) - 1:
                    target = hostname
                    target_config = {"port": cfg["port"]}
                else:
                    target, target_config = self.derive_shorthand(ssh_config, hops[i + 1])

                sock = c.gateway(hostname, cfg["port"], target, target_config["port"])
            cfg["sock"] = sock

        return hostname, cfg, forward_agent, missing_host_key_policy, host_keys_file

    @staticmethod
    def derive_shorthand(ssh_config, host_string):
        shorthand_config = {}
        user_hostport = host_string.rsplit("@", 1)
        hostport = user_hostport.pop()
        user = user_hostport[0] if user_hostport and user_hostport[0] else None
        if user:
            shorthand_config["username"] = user

        # IPv6: can't reliably tell where addr ends and port begins, so don't
        # try (and don't bother adding special syntax either, user should avoid
        # this situation by using port=).
        if hostport.count(":") > 1:
            hostname = hostport
        # IPv4: can split on ':' reliably.
        else:
            host_port = hostport.rsplit(":", 1)
            hostname = host_port.pop(0) or None
            if host_port and host_port[0]:
                shorthand_config["port"] = int(host_port[0])

        base_config = ssh_config.lookup(hostname)

        config = {
            "port": base_config.get("port", 22),
            "username": base_config.get("user"),
        }
        config.update(shorthand_config)

        return hostname, config



"""
Provides operations to set SELinux file contexts, booleans and port types.
"""
from pyinfra import host
from pyinfra.api import QuoteString, StringCommand, operation
from pyinfra.facts.selinux import FileContext, FileContextMapping, SEBoolean, SEPort, SEPorts
from pyinfra.facts.server import Which


@operation(
    pipeline_facts={"seboolean": "bool_name"},
)
def boolean(bool_name, value, persistent=False):
    """
    Set the specified SELinux boolean to the desired state.

    + boolean: name of an SELinux boolean
    + state: 'on' or 'off'
    + persistent: whether to write updated policy or not

    Note: This operation requires root privileges.

    **Example:**

    .. code:: python

        selinux.boolean(
            name='Allow Apache to connect to LDAP server',
            'httpd_can_network_connect',
            'on',
            persistent=True
        )
    """
    _valid_states = ["on", "off"]

    if value not in _valid_states:
        raise ValueError(
            f'\'value\' must be one of \'{",".join(_valid_states)}\' but found \'{value}\'',
        )

    if host.get_fact(SEBoolean, boolean=bool_name) != value:
        persist = "-P " if persistent else ""
        yield StringCommand("setsebool", f"{persist}{bool_name}", value)
        host.create_fact(SEBoolean, kwargs={"boolean": bool_name}, data=value)
    else:
        host.noop(f"boolean '{bool_name}' already had the value '{value}'")


@operation(
    pipeline_facts={"filecontext": "path"},
)
def file_context(path, se_type):
    """
    Set the SELinux type for the specified path to the specified value.

    + path: the target path (expression) for the context
    + se_type: the SELinux type for the given target

    **Example:**

    .. code:: python

        selinux.file_context(
            name='Allow /foo/bar to be served by the web server',
            '/foo/bar',
            'httpd_sys_content_t'
        )
    """

    current = host.get_fact(FileContext, path=path) or {}
    if se_type != current.get("type", ""):
        yield StringCommand("chcon", "-t", se_type, QuoteString(path))
        host.create_fact(
            FileContext,
            kwargs={"path": path},
            data=dict(current, **{"type": se_type}),
        )
    else:
        host.noop(f"file_context: '{path}' already had type '{se_type}'")


@operation(
    pipeline_facts={"filecontextmapping": "target"},
)
def file_context_mapping(target, se_type=None, present=True):
    """
    Set the SELinux file context mapping for paths matching the target.

    + target: the target path (expression) for the context
    + se_type: the SELinux type for the given target
    + present: whether to add or remove the target -> context mapping

    Note: `file_context` does not change the SELinux file context for existing files
    so `restorecon` may need to be run manually if the file contexts cannot be created
    before the related files.

    **Example:**

    .. code:: python

        selinux.file_context_mapping(
            name='Allow Apache to serve content from the /web directory',
            r'/web(/.*)?',
            se_type='httpd_sys_content_t'
        )
    """
    if present and (se_type is None):
        raise ValueError("se_type must have a valid value if present is set")

    current = host.get_fact(FileContextMapping, target=target)
    kwargs = {"target": target}
    if present:
        option = "-a" if len(current) == 0 else ("-m" if current.get("type") != se_type else "")
        if option != "":
            yield StringCommand("semanage", "fcontext", option, "-t", se_type, QuoteString(target))
            host.create_fact(
                FileContextMapping,
                kwargs=kwargs,
                data=dict(current, **{"type": se_type}),
            )
        else:
            host.noop(f"mapping for '{target}' -> '{se_type}' already present")
    else:
        if len(current) > 0:
            yield StringCommand("semanage", "fcontext", "-d", QuoteString(target))
            host.create_fact(FileContextMapping, kwargs=kwargs, data={})
        else:
            host.noop(f"no existing mapping for '{target}'")


@operation(
    pipeline_facts={"which": "sepolicy"},
)
def port(protocol, port_num, se_type=None, present=True):
    """
    Set the SELinux type for the specified protocol and port.

    + protocol: the protocol: (udp|tcp|sctp|dccp)
    + port: the port
    + se_type: the SELinux type for the given port
    + present: whether to add or remove the SELinux type for the port

    Note: This operation requires root privileges.

    **Example:**

    .. code:: python

        selinux.port(
            name='Allow Apache to provide service on port 2222',
            'tcp',
            2222,
            'http_port_t',
        )
    """

    if present and (se_type is None):
        raise ValueError("se_type must have a valid value if present is set")

    new_type = se_type if present else ""
    direct_get = len(host.get_fact(Which, command=SEPort.requires_command) or "") > 0
    if direct_get:
        current = host.get_fact(SEPort, protocol=protocol, port=port_num)
    else:
        port_info = host.get_fact(SEPorts)
        current = port_info.get(protocol, {}).get(str(port_num), "")

    if present:
        option = "-a" if current == "" else ("-m" if current != se_type else "")
        if option != "":
            yield StringCommand("semanage", "port", option, "-t", se_type, "-p", protocol, port_num)
        else:
            host.noop(f"setype for '{protocol}/{port_num}' is already '{se_type}'")
    else:
        if current != "":
            yield StringCommand("semanage", "port", "-d", "-p", protocol, port_num)
        else:
            host.noop(f"setype for '{protocol}/{port_num}' is already unset")

    if (present and (option != "")) or (not present and (current != "")):
        if direct_get:
            host.create_fact(SEPort, kwargs={"protocol": protocol, "port": port_num}, data=new_type)
        else:
            if protocol not in port_info:
                port_info[protocol] = {}
            port_info[protocol][str(port_num)] = new_type



import re


def parse_packages(regex, output):
    packages = {}

    for line in output:
        matches = re.match(regex, line)

        if matches:
            name = matches.group(1)
            packages.setdefault(name, set())
            packages[name].add(matches.group(2))

    return packages


def _parse_yum_or_zypper_repositories(output):
    repos = []

    current_repo = {}
    for line in output:
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        if line.startswith("["):
            if current_repo:
                repos.append(current_repo)
                current_repo = {}

            current_repo["name"] = line[1:-1]

        if current_repo and "=" in line:
            key, value = line.split("=", 1)
            current_repo[key] = value

    if current_repo:
        repos.append(current_repo)

    return repos


parse_yum_repositories = _parse_yum_or_zypper_repositories

parse_zypper_repositories = _parse_yum_or_zypper_repositories



"""
The ``@vagrant`` connector reads the current Vagrant status and generates an
inventory for any running VMs.

.. code:: shell

    # Run on all hosts
    pyinfra @vagrant ...

    # Run on a specific VM
    pyinfra @vagrant/my-vm-name ...

    # Run on multiple named VMs
    pyinfra @vagrant/my-vm-name,@vagrant/another-vm-name ...
"""

import json
from os import path
from queue import Queue
from threading import Thread

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize
from pyinfra.progress import progress_spinner


def _get_vagrant_ssh_config(queue, progress, target):
    logger.debug("Loading SSH config for %s", target)

    queue.put(
        local.shell(
            "vagrant ssh-config {0}".format(target),
            splitlines=True,
        ),
    )

    progress(target)


@memoize
def get_vagrant_config(limit=None):
    logger.info("Getting Vagrant config...")

    if limit and not isinstance(limit, (list, tuple)):
        limit = [limit]

    with progress_spinner({"vagrant status"}) as progress:
        output = local.shell(
            "vagrant status --machine-readable",
            splitlines=True,
        )
        progress("vagrant status")

    targets = []

    for line in output:
        line = line.strip()
        _, target, type_, data = line.split(",", 3)

        # Skip anything not in the limit
        if limit is not None and target not in limit:
            continue

        # For each running container - fetch it's SSH config in a thread - this
        # is because Vagrant *really* slow to run each command.
        if type_ == "state" and data == "running":
            targets.append(target)

    threads = []
    config_queue = Queue()

    with progress_spinner(targets) as progress:
        for target in targets:
            thread = Thread(
                target=_get_vagrant_ssh_config,
                args=(config_queue, progress, target),
            )
            threads.append(thread)
            thread.start()

    for thread in threads:
        thread.join()

    queue_items = list(config_queue.queue)

    lines = []
    for output in queue_items:
        lines.extend([ln.strip() for ln in output])

    return lines


@memoize
def get_vagrant_options():
    if path.exists("@vagrant.json"):
        with open("@vagrant.json", "r", encoding="utf-8") as f:
            return json.loads(f.read())
    return {}


def _make_name_data(host):
    vagrant_options = get_vagrant_options()
    vagrant_host = host["Host"]

    data = {
        "ssh_hostname": host["HostName"],
    }

    for config_key, data_key in (
        ("Port", "ssh_port"),
        ("User", "ssh_user"),
        ("IdentityFile", "ssh_key"),
    ):
        if config_key in host:
            data[data_key] = host[config_key]

    # Update any configured JSON data
    if vagrant_host in vagrant_options.get("data", {}):
        data.update(vagrant_options["data"][vagrant_host])

    # Work out groups
    groups = vagrant_options.get("groups", {}).get(vagrant_host, [])

    if "@vagrant" not in groups:
        groups.append("@vagrant")

    return "@vagrant/{0}".format(host["Host"]), data, groups


def make_names_data(limit=None):
    vagrant_ssh_info = get_vagrant_config(limit)

    logger.debug("Got Vagrant SSH info: \n%s", vagrant_ssh_info)

    hosts = []
    current_host = None

    for line in vagrant_ssh_info:
        # Vagrant outputs an empty line between each host
        if not line:
            if current_host:
                hosts.append(_make_name_data(current_host))

            current_host = None
            continue

        key, value = line.split(" ", 1)

        if key == "Host":
            if current_host:
                hosts.append(_make_name_data(current_host))

            # Set the new host
            current_host = {
                key: value,
            }

        elif current_host:
            current_host[key] = value

        else:
            logger.debug("Extra Vagrant SSH key/value (%s=%s)", key, value)

    if current_host:
        hosts.append(_make_name_data(current_host))

    if not hosts:
        if limit:
            raise InventoryError("No running Vagrant instances matching `{0}` found!".format(limit))
        raise InventoryError("No running Vagrant instances found!")

    return hosts



"""
Manage apk packages.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.apk import ApkPackages

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def upgrade(available: bool = False):
    """
    Upgrades all apk packages.

    + available: force all packages to be upgraded (recommended on whole Alpine version upgrades)
    """

    if available:
        yield "apk upgrade --available"
    else:
        yield "apk upgrade"


_upgrade = upgrade  # noqa: E305


@operation(is_idempotent=False)
def update():
    """
    Updates apk repositories.
    """

    yield "apk update"


_update = update  # noqa: E305


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    upgrade=False,
):
    """
    Add/remove/update apk packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``apk update`` before installing packages
    + upgrade: run ``apk upgrade`` before installing packages

    Versions:
        Package versions can be pinned like apk: ``<pkg>=<version>``.

    **Examples:**

    .. code:: python

        # Update package list and install packages
        apk.packages(
            name="Install Asterisk and Vim",
            packages=["asterisk", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        apk.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """

    if update:
        yield from _update()

    if upgrade:
        yield from _upgrade()

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(ApkPackages),
        present,
        install_command="apk add",
        uninstall_command="apk del",
        upgrade_command="apk upgrade",
        version_join="=",
        latest=latest,
    )



"""
The ``@mech`` connector reads the current mech status and generates an inventory
for any running VMs.

.. code:: python

    # Run on all hosts
    pyinfra @mech ...

    # Run on a specific VM
    pyinfra @mech/my-vm-name ...

    # Run on multiple named VMs
    pyinfra @mech/my-vm-name,@mech/another-vm-name ...
"""

import json
from os import path
from queue import Queue
from threading import Thread

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize
from pyinfra.progress import progress_spinner


def _get_mech_ssh_config(queue, progress, target):
    logger.debug("Loading SSH config for %s", target)

    # Note: We have to work-around the fact that "mech ssh-config somehost"
    # does not return the correct "Host" value. When "mech" fixes this
    # issue we can simply this code.
    lines = local.shell(
        "mech ssh-config {0}".format(target),
        splitlines=True,
    )

    newlines = []
    for line in lines:
        if line.startswith("Host "):
            newlines.append("Host " + target)
        else:
            newlines.append(line)

    queue.put(newlines)

    progress(target)


@memoize
def get_mech_config(limit=None):
    logger.info("Getting Mech config...")

    if limit and not isinstance(limit, (list, tuple)):
        limit = [limit]

    # Note: There is no "--machine-readable" option to 'mech status'
    with progress_spinner({"mech ls"}) as progress:
        output = local.shell(
            "mech ls",
            splitlines=True,
        )
        progress("mech ls")

    targets = []

    for line in output:

        address = ""

        data = line.split()
        target = data[0]

        if len(data) == 5:
            address = data[1]

        # Skip anything not in the limit
        if limit is not None and target not in limit:
            continue

        # For each vm that has an address, fetch it's SSH config in a thread
        if address != "" and address[0].isdigit():
            targets.append(target)

    threads = []
    config_queue = Queue()

    with progress_spinner(targets) as progress:
        for target in targets:
            thread = Thread(
                target=_get_mech_ssh_config,
                args=(config_queue, progress, target),
            )
            threads.append(thread)
            thread.start()

    for thread in threads:
        thread.join()

    queue_items = list(config_queue.queue)

    lines = []
    for output in queue_items:
        lines.extend(output)

    return lines


@memoize
def get_mech_options():
    if path.exists("@mech.json"):
        with open("@mech.json", "r", encoding="utf-8") as f:
            return json.loads(f.read())
    return {}


def _make_name_data(host):
    mech_options = get_mech_options()
    mech_host = host["Host"]

    data = {
        "ssh_hostname": host["HostName"],
    }

    for config_key, data_key in (
        ("Port", "ssh_port"),
        ("User", "ssh_user"),
        ("IdentityFile", "ssh_key"),
    ):
        if config_key in host:
            data[data_key] = host[config_key]

    # Update any configured JSON data
    if mech_host in mech_options.get("data", {}):
        data.update(mech_options["data"][mech_host])

    # Work out groups
    groups = mech_options.get("groups", {}).get(mech_host, [])

    if "@mech" not in groups:
        groups.append("@mech")

    return "@mech/{0}".format(host["Host"]), data, groups


def make_names_data(limit=None):
    mech_ssh_info = get_mech_config(limit)

    logger.debug("Got Mech SSH info: \n%s", mech_ssh_info)

    hosts = []
    current_host = None

    for line in mech_ssh_info:
        if not line:
            if current_host:
                hosts.append(_make_name_data(current_host))

            current_host = None
            continue

        key, value = line.strip().split(" ", 1)

        if key == "Host":
            if current_host:
                hosts.append(_make_name_data(current_host))

            # Set the new host
            current_host = {
                key: value,
            }

        elif current_host:
            current_host[key] = value

        else:
            logger.debug("Extra Mech SSH key/value (%s=%s)", key, value)

    if current_host:
        hosts.append(_make_name_data(current_host))

    if not hosts:
        raise InventoryError("No running Mech instances found!")

    return hosts



import json
import platform
import re
import sys
from typing import TYPE_CHECKING, Callable, Dict, Iterator, List, Tuple, Union

import click

from pyinfra import __version__, logger
from pyinfra.api.host import Host

from .util import json_encode

if TYPE_CHECKING:
    from pyinfra.api.state import State


ANSI_RE = re.compile(r"\033\[((?:\d|;)*)([a-zA-Z])")


def _strip_ansi(value):
    return ANSI_RE.sub("", value)


def _get_group_combinations(inventory: Iterator[Host]):
    group_combinations: Dict[Tuple, List[Host]] = {}

    for host in inventory:
        # Tuple for hashability, set to normalise order
        host_groups = tuple(set(host.groups))

        group_combinations.setdefault(host_groups, [])
        group_combinations[host_groups].append(host)

    return group_combinations


def _stringify_host_keys(data):
    if isinstance(data, dict):
        return {
            key.name if isinstance(key, Host) else key: _stringify_host_keys(value)
            for key, value in data.items()
        }

    return data


def jsonify(data, *args, **kwargs):
    data = _stringify_host_keys(data)
    return json.dumps(data, *args, **kwargs)


def print_state_facts(state: "State"):
    click.echo(err=True)
    click.echo("--> Facts:", err=True)
    click.echo(jsonify(state.facts, indent=4, default=json_encode), err=True)


def print_state_operations(state: "State"):
    state_ops = {host: ops for host, ops in state.ops.items() if state.is_host_in_limit(host)}

    click.echo(err=True)
    click.echo("--> Operations:", err=True)
    click.echo(jsonify(state_ops, indent=4, default=json_encode), err=True)
    click.echo(err=True)
    click.echo("--> Operation meta:", err=True)
    click.echo(jsonify(state.op_meta, indent=4, default=json_encode), err=True)

    click.echo(err=True)
    click.echo("--> Operation order:", err=True)
    click.echo(err=True)
    for op_hash in state.get_op_order():
        meta = state.op_meta[op_hash]
        hosts = set(host for host, operations in state.ops.items() if op_hash in operations)

        click.echo(
            "    {0} (names={1}, hosts={2})".format(
                op_hash,
                meta["names"],
                hosts,
            ),
            err=True,
        )


def print_groups_by_comparison(print_items, comparator=lambda item: item[0]):
    items = []
    last_name = None

    for name in print_items:
        # Keep all facts with the same first character on one line
        if last_name is None or comparator(last_name) == comparator(name):
            items.append(name)

        else:
            click.echo(
                "    {0}".format(", ".join((click.style(name, bold=True) for name in items))),
                err=True,
            )

            items = [name]

        last_name = name

    if items:
        click.echo(
            "    {0}".format(", ".join((click.style(name, bold=True) for name in items))),
            err=True,
        )


def print_fact(fact_data):
    click.echo(jsonify(fact_data, indent=4, default=json_encode), err=True)


def print_inventory(state: "State"):
    for host in state.inventory:
        click.echo(err=True)
        click.echo(host.print_prefix, err=True)
        click.echo("--> Groups: {0}".format(", ".join(host.groups)), err=True)
        click.echo("--> Data:", err=True)
        click.echo(jsonify(host.data, indent=4, default=json_encode), err=True)


def print_facts(facts):
    for name, data in facts.items():
        click.echo(err=True)
        click.echo(
            "--> Fact data for: {0}".format(
                click.style(name, bold=True),
            ),
            err=True,
        )
        print_fact(data)


def print_support_info():
    click.echo(
        """
    If you are having issues with pyinfra or wish to make feature requests, please
    check out the GitHub issues at https://github.com/Fizzadar/pyinfra/issues .
    When adding an issue, be sure to include the following:
""",
    )

    click.echo("    System: {0}".format(platform.system()), err=True)
    click.echo("      Platform: {0}".format(platform.platform()), err=True)
    click.echo("      Release: {0}".format(platform.uname()[2]), err=True)
    click.echo("      Machine: {0}".format(platform.uname()[4]), err=True)
    click.echo("    pyinfra: v{0}".format(__version__), err=True)
    click.echo("    Executable: {0}".format(sys.argv[0]), err=True)
    click.echo(
        "    Python: {0} ({1}, {2})".format(
            platform.python_version(),
            platform.python_implementation(),
            platform.python_compiler(),
        ),
        err=True,
    )


def print_rows(rows):
    # Go through the rows and work out all the widths in each column
    column_widths = []

    for _, columns in rows:
        if isinstance(columns, str):
            continue

        for i, column in enumerate(columns):
            if i >= len(column_widths):
                column_widths.append([])

            # Length of the column (with ansi codes removed)
            width = len(_strip_ansi(column.strip()))
            column_widths[i].append(width)

    # Get the max width of each column and add 4 padding spaces
    column_widths = [max(widths) + 4 for widths in column_widths]

    # Now print each column, keeping text justified to the widths above
    for func, columns in rows:
        line = columns

        if not isinstance(columns, str):
            justified = []

            for i, column in enumerate(columns):
                stripped = _strip_ansi(column)
                desired_width = column_widths[i]
                padding = desired_width - len(stripped)

                justified.append(
                    "{0}{1}".format(
                        column,
                        " ".join("" for _ in range(padding)),
                    ),
                )

            line = "".join(justified)

        func(line)


def print_meta(state: "State"):
    group_combinations = _get_group_combinations(state.inventory.iter_activated_hosts())
    rows: List[Tuple[Callable, Union[List[str], str]]] = []

    for i, (groups, hosts) in enumerate(group_combinations.items(), 1):
        if not hosts:
            continue

        if groups:
            rows.append(
                (
                    logger.info,
                    "Groups: {0}".format(
                        click.style(" / ".join(groups), bold=True),
                    ),
                ),
            )
        else:
            rows.append((logger.info, "Ungrouped:"))

        for host in hosts:
            meta = state.meta[host]

            # Didn't connect to this host?
            if host not in state.activated_hosts:
                rows.append(
                    (
                        logger.info,
                        [
                            host.style_print_prefix("red", bold=True),
                            click.style("No connection", "red"),
                        ],
                    ),
                )
                continue

            rows.append(
                (
                    logger.info,
                    [
                        host.print_prefix,
                        "Operations: {0}".format(meta["ops"]),
                        "Change: {0}".format(meta["ops_change"]),
                        "No change: {0}".format(meta["ops_no_change"]),
                    ],
                ),
            )

        if i != len(group_combinations):
            rows.append((lambda m: click.echo(m, err=True), []))

    print_rows(rows)


def print_results(state: "State"):
    group_combinations = _get_group_combinations(state.inventory.iter_activated_hosts())
    rows: List[Tuple[Callable, Union[List[str], str]]] = []

    for i, (groups, hosts) in enumerate(group_combinations.items(), 1):
        if not hosts:
            continue

        if groups:
            rows.append(
                (
                    logger.info,
                    "Groups: {0}".format(
                        click.style(" / ".join(groups), bold=True),
                    ),
                ),
            )
        else:
            rows.append((logger.info, "Ungrouped:"))

        for host in hosts:
            # Didn't connect to this host?
            if host not in state.activated_hosts:
                rows.append(
                    (
                        logger.info,
                        [
                            host.style_print_prefix("red", bold=True),
                            click.style("No connection", "red"),
                        ],
                    ),
                )
                continue

            results = state.results[host]

            meta = state.meta[host]
            success_ops = results["success_ops"]
            partial_ops = results["partial_ops"]
            # TODO: type meta object
            changed_ops = success_ops - meta["ops_no_change"]  # type: ignore
            error_ops = results["error_ops"]
            ignored_error_ops = results["ignored_error_ops"]

            host_args = ("green",)
            host_kwargs = {}

            # If all ops got complete
            if results["ops"] == meta["ops"]:
                # We had some errors - but we ignored them - so "warning" color
                if error_ops != 0:
                    host_args = ("yellow",)

            # Ops did not complete!
            else:
                host_args = ("red",)
                host_kwargs["bold"] = True

            changed_str = "Changed: {0}".format(click.style(f"{changed_ops}", bold=True))
            if partial_ops:
                changed_str = f"{changed_str} ({partial_ops} partial)"

            error_str = "Errors: {0}".format(click.style(f"{error_ops}", bold=True))
            if ignored_error_ops:
                error_str = f"{error_str} ({ignored_error_ops} ignored)"

            rows.append(
                (
                    logger.info,
                    [
                        host.style_print_prefix(*host_args, **host_kwargs),
                        changed_str,
                        "No change: {0}".format(click.style(f"{meta['ops_no_change']}", bold=True)),
                        error_str,
                    ],
                ),
            )

        if i != len(group_combinations):
            rows.append((lambda m: click.echo(m, err=True), []))

    print_rows(rows)
